/*
    Copyright (C) 1999 by  XCIN TEAM

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef  _WINLIST_H
#define  _WINLIST_H

#include <X11/Xlib.h>
#include <X11/Xft/Xft.h>
#include <X11/cursorfont.h>
#include "oximtool.h"
typedef struct gui_s gui_t;

/*
 *  GUI Window configuration.
 */
#define WTYPE_MAIN		0
#define WTYPE_GUIREQ		1
#define WTYPE_OVERSPOT		2

#define WMODE_MAP		1
#define	WMODE_EXIT		2

typedef struct winlist_s winlist_t;
struct winlist_s {
    Window window;
    unsigned int font_size;
    unsigned int num_fonts;
    XftFont **xftfonts;
    XftDraw *draw;
    xmode_t winmode;

    int pos_x, pos_y;
    unsigned int width, height;

    void *data;
    void (*win_draw_func)(gui_t *, winlist_t *);
    void (*win_event_func)(gui_t *, winlist_t *, XEvent *);
    void (*win_attrib_func)(gui_t *, winlist_t *, XConfigureEvent *, int);
    void (*win_destroy_func)(gui_t *, winlist_t *);
    winlist_t *next;
};

/*
 *  GUI global configuration.
 */
#define WIN_CHANGE_IM           0x0001
#define WIN_CHANGE_FOCUS        0x0002
#define WIN_CHANGE_REDRAW       0x000f
#define WIN_CHANGE_BELL		0x0010
#define WIN_CHANGE_BELL2	0x0020
#define WIN_CHANGE_BELLALL	0x00f0

#define WIN_MONITOR_CLIENT	0x0001
#define WIN_MONITOR_FOCUS	0x0002
#define WIN_MONITOR_OVERSPOT	0x0004

#define MAX_COLORS   24
#define NO_COLOR     -1

#define WINBOX_COLOR 0
#define BORDER_COLOR 1
#define LIGHT_COLOR  2
#define DARK_COLOR   3
#define CURSOR_COLOR 4 
#define CURSORFONT_COLOR 5 
#define FONT_COLOR   6
#define CONVERTNAME_COLOR 7
#define INPUTNAME_COLOR 8
#define UNDERLINE_COLOR 9
#define KEYSTROKE_COLOR 10
#define KEYSTROKEBOX_COLOR 11
#define SELECTFONT_COLOR 12
#define SELECTBOX_COLOR 13
#define MENU_BG_COLOR 14
#define MENU_FONT_COLOR 15

#define XCIN_BORDER_COLOR 16
#define XCIN_BG_COLOR 17
#define XCIN_FONT_COLOR 18
#define XCIN_STATUS_FG_COLOR 19
#define XCIN_STATUS_BG_COLOR 20
#define XCIN_CURSOR_FG_COLOR 21
#define XCIN_CURSOR_BG_COLOR 22
#define XCIN_UNDERLINE_COLOR 23

struct gui_s {
    Display *display;
    int display_width, display_height;
    int screen;
    Visual *visual;

    Window root;
    Atom wm_del_win;
    Colormap colormap;
    XftColor colors[MAX_COLORS];

    int xcin_style; /* 是否使用 XCIN 輸入風格 */
    winlist_t *xcin_win;

    winlist_t *root_win, *status_win, *preedit_win, *select_win, *tray_win;
    winlist_t *menu_win, *symbol_win, *keyboard_win;
    xmode_t winchange;
};

extern winlist_t *gui_search_win(Window window);
extern winlist_t *gui_new_win(void);
extern void gui_freewin(Window window);
extern void gui_winmap_change(winlist_t *win, int state);
extern int gui_check_window(Window window);
extern void gui_set_monitor(Window w, xmode_t flag, int icid);

extern void gui_move_window(gui_t *gui, winlist_t *win);
extern unsigned int gui_Draw_String(winlist_t *win, int fgcolor_idx, int bgcolor_idx, int x, int y, char *string, int len);
extern unsigned int gui_TextEscapement(winlist_t *win, char *string, int len);
extern void gui_Draw_3D_Box(winlist_t *win, int x, int y, int width, int height, int fill_idx, int boxstyle);
/* 劃線 */
extern void gui_Draw_Line(winlist_t *win, int x, int y, int x1, int y1, int color_idx, int style);

extern unsigned long gui_Color(int index);
extern unsigned int gui_FontSize(winlist_t *win);
extern void gui_Draw_Image(winlist_t *win, int x, int y, char **xpm_data);
/* 送出模擬按鍵到指定的視窗 */
extern void gui_send_key(Window win, int state, KeySym keysym);
extern void split_lc_ctype(char *loc_buf, char *loc, char *enc);
/* 取得 Focus Window (None : 沒有) */
extern Window gui_get_input_focus(void);
/* 取得 LED 狀態 */
extern int gui_get_led_state(KeySym keysym);

extern Cursor gui_cursor_hand;
extern Cursor gui_cursor_move;
extern Cursor gui_cursor_arrow;
extern int gui_menu_actived(void);
extern int gui_show_menu(gui_t *gui, winlist_t *parent_win);
extern int gui_hide_menu(gui_t *gui);

/* 切換 XCIN 輸入風格 */
extern void gui_xcin_style_switch(gui_t *gui);

extern void gui_switch_symbol(gui_t *gui);
extern int gui_symbol_actived(void);
extern int gui_show_symbol(gui_t *gui);
extern int gui_hide_symbol(gui_t *gui);

extern void gui_switch_keyboard(gui_t *gui);
extern int gui_keyboard_actived(void);
extern void gui_show_keyboard(gui_t *gui);
extern void gui_hide_keyboard(gui_t *gui);
#endif
