/*
    Copyright (C) 1999 by  XCIN TEAM

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/      


#ifndef _OXIMTOOL_H
#define _OXIMTOOL_H

#include <stdio.h>
#include <stdlib.h>

#ifdef __cplusplus
extern "C" {
#endif

/* Global status variables. */
extern int verbose, errstatus;

#if !defined(True) || !defined(False)
#undef  True
#undef  False
#define True  1
#define False 0
#endif

/* Integer types. */
typedef unsigned int	xmode_t;
typedef signed char	x_int8;
typedef unsigned char	x_uint8;

#if (SIZEOF_SHORT == 2)
typedef short		x_int16;
typedef unsigned short	x_uint16;
#else
#if (SIZEOF_INT == 2)
typedef int		x_int16;
typedef unsigned int	x_uint16;
#endif
#endif

#if (SIZEOF_INT == 4)
typedef int		x_int32;
typedef unsigned int	x_uint32;
#else
#if (SIZEOF_LONG == 4)
typedef long		x_int32;
typedef unsigned long	x_uint32;
#else
#if (SIZEOF_SHORT == 4)
typedef short		x_int32;
typedef unsigned short	x_uint32;
#endif
#endif
#endif

/* For general message level. */
#define OXIMMSG_NORMAL		 0		/* normal		*/
#define OXIMMSG_WARNING		 1		/* warning		*/
#define OXIMMSG_IWARNING	 2		/* internal warnning	*/
#define OXIMMSG_ERROR		-1		/* error		*/
#define OXIMMSG_IERROR		-2		/* internal error	*/
#define OXIMMSG_EMPTY		 3		/* pure message printed */

/* For international message output (gettext) */
#ifdef HAVE_LIBINTL_H
#  include <libintl.h>
#  define _(STRING) gettext(STRING)
#else
#  define _(STRING) STRING
#endif
#ifndef N_
#  define N_(STRING) STRING
#endif


#define N_ASCII                 95
#define UINT32BIT               0x80000000

/* File type for check_file_exist(); */
enum ftype {
    FTYPE_FILE,			/* Regular file */
    FTYPE_DIR,			/* Directory */
    FTYPE_NONE
};

/* Input data type for set_data(); */
enum rctype {
    RC_BFLAG,
    RC_SFLAG,
    RC_IFLAG,
    RC_LFLAG,
    RC_BYTE,
    RC_UBYTE,
    RC_SHORT,
    RC_USHORT,
    RC_INT,
    RC_UINT,
    RC_LONG,
    RC_ULONG,
    RC_FLOAT,
    RC_DOUBLE,
    RC_STRING,
    RC_STRARR,
    RC_NONE
};

/* OXIM global resources. */
typedef struct {
    char *lc_ctype;
    char *lc_messages;
    char *locale;
    char *encoding;
    int encid, locid;
} locale_t;

typedef struct {
    int argc;			/* Command line arguement list */
    char **argv;
    locale_t locale;            /* Locale name. */
    char *usrhome;		/* User home directory. */
    char *default_dir;          /* Default module directory. */
    char *user_dir;             /* User data directory. */
    char *rcfile;               /* rcfile name. */
    void *rclist;		/* resource list read from the rcfile */
} oxim_rc_t;

/* 
 * General module definition.
 */
enum mtype {
    MTYPE_IM			/* IM module */
};

#define MODULE_ID_SIZE 20	/* module id buffer size */
typedef struct {		/* common module header */
    enum mtype module_type;
    char *name;
    char *version;
    char *comments;
} mod_header_t;


/* 
 * Replacement of the standard libc functions.
 */
#ifdef HAVE_MERGESORT
#  define stable_sort  mergesort
#else
#  define stable_sort  oxim_mergesort
   extern void oxim_mergesort(void *base, size_t nmemb, size_t size,
            int (*compar)(const void *, const void *));
#endif

#ifndef HAVE_SNPRINTF
#  define snprintf oxim_snprintf
   extern int oxim_snprintf(char *str, size_t size, const char *format, ...);
#endif

#define DebugLog(defverb, action) \
	if (defverb <= verbose) perr_debug action

/* oximtool functions. */
extern void set_perr(char *error_head);
extern void perr(int exitcode, const char *fmt, ...);
extern void perr_debug(const char *fmt, ...);
extern void *oxim_malloc(size_t n_bytes, int reset);
extern void *oxim_realloc(void *pt, size_t n_bytes);
extern int set_lc_ctype(char *loc_name, char *loc_return, int loc_size,
		char *enc_return, int enc_size, int exitcode);
extern int set_lc_messages(char *loc_name, char *loc_return, int loc_size);
extern int set_lc_ctype_env(char *loc_name, char *loc_return, int loc_size, 
		char *enc_return, int enc_size, int exitcode);

extern FILE *open_file(char *fn, char *md, int exitcode);
extern int check_file_exist(char *path, int type);
extern int check_datafile(char *fn, char *sub_path,
		oxim_rc_t *xrc, char *true_fn, int true_fnsize);
extern void check_oxim_path(oxim_rc_t *xrc, int exitcode);
extern mod_header_t *load_module(char *modname, enum mtype mod_type,
		char *version, oxim_rc_t *xrc, char *sub_path);
extern void unload_module(mod_header_t *imodule);
extern void module_comment(mod_header_t *modp, char *mod_name);

extern void read_oximrc(oxim_rc_t *xrc, char *rcfile);
extern int get_resource(oxim_rc_t *xrc, char **cmd_list, char *value,
		int v_size, int n_cmd_list);
extern int get_resource_long(oxim_rc_t *xrc, char **cmd_list, char *value,
		int v_size, int n_cmd_list, int sep);

extern int get_line(char *str, int str_size, FILE *f, int *lineno, 
		char *ignore_ch);
extern int get_word(char **line, char *word, int word_size, char *token);
extern void set_data(void *ref, int type, char *value, 
		unsigned long flag_mask, int bufsize);
extern int strcmp_wild(char *s1, char *s2);

extern int oxim_newenc(char *encoding, oxim_rc_t *xrc);
extern int oxim_enc2id(char *encoding);
extern char *oxim_id2enc(int encid);
extern void oxim_enclist_clean(void);
extern int connect_iconv(int encid, int to_wch);

/*
 * The following is absolute .....
 */

/* General char type: mbs encoding
 *
 * Note: In Linux, if uch_t.s = "a1 a4", then uch_t.uch = 0xa4a1, i.e.,
 *       the order reversed. This might not be the general case for all
 *       plateforms.
 */
#ifndef UCH_SIZE
#define uchar_t long long
#define UCH_SIZE sizeof(uchar_t) 
typedef union { 
    unsigned char s[UCH_SIZE];
    uchar_t uch;
} uch_t;
#endif


typedef signed char byte_t;
typedef unsigned char ubyte_t;
typedef unsigned short xtype_t;

extern int wchs_to_mbs(char *mbs, uch_t *wchs, int size);
extern int nwchs_to_mbs(char *mbs, uch_t *wchs, int n_wchs, int size);
extern int oxim_utf8_to_ucs4(const char *src_orig, unsigned int *ucs4, int len);
extern int oxim_ucs4_to_utf8(const unsigned int ucs4, char *dst);
extern char *oxim_output_filter(char *input, int is_t2s);

#ifdef __cplusplus 
} /* extern "C" */
#endif 

#endif
