/*
    Copyright (C) 1999 by  XCIN TEAM

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/      


#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <zlib.h>
#include "oximtool.h"
#include "imodule.h"
#include "gencin.h"

static char *default_module_name = "gen-inp.so";
static int im_idx;

/*----------------------------------------------------------------------------

	Dynamic module loading:

	For ELF systems (Solaris, GNU/Linux, FreeBSD, HP-UX.11)
	For HP-UX PA32, PA64 systems.

----------------------------------------------------------------------------*/

#ifdef	HAVE_DLOPEN			/* ELF */
#include <dlfcn.h>
#else
#ifdef	HAVE_SHL_LOAD			/* HP-UX */
#include <dl.h>
#endif
#endif

#define BUFLEN	1024

struct _mod_stack_s {
   void *ld;
   mod_header_t *modp;
   int ref;
   struct _mod_stack_s *next;
};

struct _mod_stack_s *mod_stack = NULL;

/*----------------------------------------------------------------------------

	Load and Unload modules.

----------------------------------------------------------------------------*/

static int
find_module(char *path, int path_size, oxim_rc_t *xrc, char *sub_path)
{
    char fn[BUFLEN];

    if (check_datafile(path, sub_path, xrc, fn, BUFLEN) == False)
	return False;

    strcpy(path, fn);
    return check_file_exist(path, FTYPE_FILE);
}

mod_header_t *
load_module(char *modname, enum mtype mod_type, char *version,
	    oxim_rc_t *xrc, char *sub_path)
{
    struct _mod_stack_s *ms=mod_stack;
    char ldfn[BUFLEN];
    mod_header_t *module;
    void *ld=NULL;
    int err=1;

    while (ms != NULL) {
	if (strcmp(modname, ms->modp->name) == 0) {
	    ms->ref ++;
	    return ms->modp;
	}
	ms = ms->next;
    }

    snprintf(ldfn, BUFLEN, "%s.so", modname);
    if (find_module(ldfn, BUFLEN, xrc, sub_path) == True &&
#ifdef	HAVE_DLOPEN
        (ld = dlopen(ldfn, RTLD_LAZY)) != NULL)
#else
#ifdef	HAVE_SHL_LOAD
        (ld = (void *)shl_load(ldfn, BIND_DEFERRED, 0L)) != NULL)
#endif
#endif
	err = 0;

    if (err) {
#ifdef	HAVE_DLOPEN
	char *errstr = dlerror();
        perr(OXIMMSG_IWARNING, N_("dlerror: %s\n"), errstr);
#endif
	ld = NULL;
    }

#ifdef	HAVE_DLOPEN
    if (!err && !(module = (mod_header_t *)dlsym(ld, "module_ptr")))
#else
#ifdef	HAVE_SHL_LOAD
    if (!err && shl_findsym((shl_t *)&ld, "module_ptr", TYPE_DATA, &module)!=0)
#endif
#endif
    {
	perr(OXIMMSG_IWARNING, N_("module symbol \"module_ptr\" not found.\n"));
	err = 1;
    }
    if (!err && module->module_type != mod_type) {
	perr(OXIMMSG_IWARNING,
	     N_("invalid module type, type %d required.\n"), mod_type);
	err = 1;
    }
    if (!err && strcmp(module->version, version) != 0) {
	perr(OXIMMSG_IWARNING,
	     N_("invalid module version: %s, version %s required.\n"),
	     module->version, version);
    }

    if (err) {
        perr(OXIMMSG_WARNING,
	     N_("cannot load module \"%s\", ignore.\n"), modname);
	if (ld)
#ifdef HAVE_DLOPEN
	    dlclose(ld);
#else
#ifdef HAVE_SHL_OPEN
	    shl_unload((shl_t)ld);
#endif
#endif
	return NULL;
    }
    else {
	ms = oxim_malloc(sizeof(struct _mod_stack_s), 0);
	ms->ld = ld;
	ms->modp = module;
	ms->ref = 1;
	ms->next = mod_stack;
	mod_stack = ms;
	return module;
    }
}

void
unload_module(mod_header_t *modp)
{
    struct _mod_stack_s *ms = mod_stack;

    while (ms != NULL) {
	if (modp == ms->modp) {
	    ms->ref --;
	    break;
	}
	ms = ms->next;
    }
    if (ms && ms->ref <= 0) {
#ifdef	HAVE_DLOPEN
	dlclose(ms->ld);
#else
#ifdef	HAVE_SHL_LOAD
	shl_unload((shl_t)(ms->ld));
#endif
#endif
	mod_stack = ms->next;
	free(ms);
    }
}

/*----------------------------------------------------------------------------

	Print the module comment.

----------------------------------------------------------------------------*/

void
module_comment(mod_header_t *modp, char *mod_name)
{
    if (modp) {
	perr(OXIMMSG_NORMAL, N_("module \"%s\":"), mod_name);
	if (modp->comments)
	    perr(OXIMMSG_EMPTY, "\n\n%s\n", N_(modp->comments));
	else
	    perr(OXIMMSG_EMPTY, N_("no comments available.\n"));
    }
}

#include "imodule.h"

static int NumberOfIM = 0;
static im_t **IMList = NULL;

/*---------------------------------------------------------------------------

        Load Module & Module Init.

---------------------------------------------------------------------------*/

static int
check_module(module_t *modp, char *objname)
{
    char *str = NULL, **objn;
    int check_ok=0;

    if (modp == NULL)
	return False;
/*
 *  Check for necessary symbols.
 */
    if (! modp->conf_size)
	str = "conf_size";
    else if (! modp->init)
	str = "init";
    else if (! modp->xim_init)
	str = "xim_init";
    else if (! modp->xim_end)
	str = "xim_end";
    else if (! modp->keystroke)
	str = "keystroke";
    else if (! modp->show_keystroke)
	str = "show_keystroke";
    if (str) {
	perr(OXIMMSG_IWARNING,
	    N_("undefined symbol \"%s\" in module \"%s\", ignore.\n"),
	    str, modp->module_header.name);
	return  False;
    }
/*
 *  Check for the valid objname.
 */
    objn = modp->valid_objname;
    if (!objn) {
	if (strcmp_wild(modp->module_header.name, objname) == 0)
	    check_ok = 1;
    }
    else {
	while (*objn) {
	    if(strcmp_wild(*objn, objname) == 0) {
		check_ok = 1;
		break;
	    }
	    objn ++;
	}
    }
    if (check_ok == 0) {
	perr(OXIMMSG_WARNING,
	    N_("invalid objname \"%s\" for module \"%s\", ignore.\n"),
	    objname, modp->module_header.name);
	return False;
    }
    return True;
}

static imodule_t *
creat_module(module_t *templet, char *objname)
{
    imodule_t *imodp;

    imodp = oxim_malloc(sizeof(imodule_t), 1);
    imodp->modp = (void *)templet;
    imodp->name = templet->module_header.name;
    imodp->version = templet->module_header.version;
    imodp->comments = templet->module_header.comments;
    imodp->module_type = templet->module_header.module_type;
    imodp->conf = oxim_malloc(templet->conf_size, 1);
    imodp->init = templet->init;
    imodp->xim_init = templet->xim_init;
    imodp->xim_end = templet->xim_end;
    imodp->keystroke = templet->keystroke;
    imodp->show_keystroke = templet->show_keystroke;
    imodp->terminate = templet->terminate;

    imodp->objname = (objname) ? (char *)strdup(objname) : imodp->name; 

    return imodp;
}

static imodule_t *
IM_load(char *modname, char *objname, oxim_rc_t *xrc)
{
    module_t *modp;
    imodule_t *imodp;
    int load_ok=1;

    modp = (module_t *)load_module(modname,MTYPE_IM,MODULE_VERSION,xrc, "modules");
    if (check_module(modp, objname) == False)
	load_ok = 0;
    else {
	imodp = creat_module(modp, objname);
	if (imodp->init(imodp->conf, objname, xrc) != True) {
	    free(imodp->conf);
	    free(imodp);
	    load_ok = 0;
	}
    }
    if (load_ok == 0) {
	perr(OXIMMSG_WARNING,
	    N_("cannot load IM: %s, ignore.\n"), objname);
	unload_module((mod_header_t *)modp);
	return NULL;
    }
    return imodp;
}

/*----------------------------------------------------------------------------

	Cinput structer manager.

----------------------------------------------------------------------------*/
static void OXIM_IMFree(int idx)
{
    if (idx < 0 || idx >= NumberOfIM)
	return;

    im_t *imp;

    imp = IMList[idx];
    if (imp->inpname)
    {
	free(imp->inpname);
	imp->inpname = NULL;
    }
    if (imp->modname)
    {
	free(imp->modname);
	imp->modname = NULL;
    }
    if (imp->objname)
    {
	free(imp->objname);
	imp->objname = NULL;
    }
    if (imp->imodp)
    {
	if (imp->imodp->terminate)
	    imp->imodp->terminate(imp->imodp->conf);
	if (imp->imodp->modp)
	    unload_module((mod_header_t *)imp->imodp->modp);
	if (imp->imodp->conf)
	    free(imp->imodp->conf);
	free(imp->imodp);
	imp->imodp = NULL;
    }
    imp->key = -1;
}

imodule_t *
OXIM_IMGet(int idx, oxim_rc_t *xrc)
{
    if (idx < 0 || idx >= NumberOfIM)
	return NULL;

    im_t *imp = IMList[idx];

    if (imp->modname && imp->objname && imp->imodp == NULL)
	imp->imodp = IM_load(imp->modname, imp->objname, xrc);

    if (imp->imodp == NULL)
    {
	OXIM_IMFree(idx);
    }

    return imp->imodp;
}

imodule_t *
OXIM_IMGetNext(int idx, int *idx_ret, oxim_rc_t *xrc)
{
    int i, j;
    im_t *imp;

    if (idx < 0 || idx >= NumberOfIM)
	return NULL;

    *idx_ret = -1;

    for (j = 0, i = idx ; j < NumberOfIM ; j++, i++)
    {
	if (i >= NumberOfIM)
	{
	    i = 0;
	}
	imp = IMList[i];

	if (imp->modname && imp->objname)
	{
	    if (imp->imodp == NULL)
	    {
		imp->imodp = IM_load(imp->modname, imp->objname, xrc);
		if (imp->imodp != NULL )
		{
		    *idx_ret = i;
		    break;
		}
	    }
	}
    }
    if (imp->imodp == NULL)
	OXIM_IMFree(*idx_ret);
    return imp->imodp;
}

imodule_t *
OXIM_IMGetPrev(int idx, int *idx_ret, oxim_rc_t *xrc)
{
    int i, j;
    im_t *imp;

    if (idx < 0 || idx >= NumberOfIM)
	return NULL;

    *idx_ret = -1;

    for (j = 0, i = idx ; j < NumberOfIM ; j++, i--)
    {
	if (i < 0)
	{
	    i = NumberOfIM - 1;
	}
	imp = IMList[i];

	if (imp->modname && imp->objname) {
	    *idx_ret = i;
	    break;
	}
    }
    if (*idx_ret != -1 && imp->modname && imp->objname && imp->imodp == NULL)
	imp->imodp = IM_load(imp->modname, imp->objname, xrc);
    if (imp->imodp == NULL)
	OXIM_IMFree(*idx_ret);
    return imp->imodp;
}

numlist_t *
IM_get_numlist(void)
{
    static numlist_t *numlist = NULL;
    if (!numlist)
    {
	free(numlist);
	numlist = oxim_malloc(sizeof(numlist) * NumberOfIM+1, True);
    }
    numlist_t *s=numlist;
    int i;

    for (i = 0 ; i < NumberOfIM ; i++)
    {
	if (IMList[i]->key >= 0 && IMList[i]->modname && IMList[i]->objname)
	{
	    *s = (numlist_t)IMList[i]->key;
	    s++;
	}
    }
    *s = (numlist_t)-1;

    return numlist;
}

imodule_t *
OXIM_IMSearch(char *objname, int *idx_ret, oxim_rc_t *xrc)
{
    int i;
    im_t *imp;

    *idx_ret = -1;
    for (i = 0 ; i < NumberOfIM; i++)
    {
	imp = IMList[i];
	if (imp->objname && strcmp(imp->objname, objname) == 0)
	{
	    *idx_ret = i;
	    break;
	}
    }

    if (*idx_ret != -1 && imp->modname && imp->objname && imp->imodp == NULL)
	imp->imodp = IM_load(imp->modname, imp->objname, xrc);
    if (imp->imodp == NULL)
	OXIM_IMFree(*idx_ret);
    return imp->imodp;
}

void
OXIM_IMFreeAll(void)
{
    if (!IMList)
	return;
 
    int i;
    for (i = 0; i < NumberOfIM ; i++)
    {
	OXIM_IMFree(i);
	free(IMList[i]);
    }

    free(IMList);
    NumberOfIM = 0;
    IMList = NULL;
}

/*-------------------------------------------------------------------
  掃描使用者自訂之輸入法表格與輸入法模組
  順序為 使用者 > 系統
  Add By Firefly(firefly@opendesktop.org.tw)
-------------------------------------------------------------------*/
void OXIM_IMRegister(oxim_rc_t *xrc, char *modname, char *objname, char *cname)
{
    int i;
    im_t *im = NULL;

    for (i = 0; i < NumberOfIM ; i++)
    {
	/* 找出第一個可用的位置 */
	if (im == NULL && !IMList[i]->objname)
	    im = IMList[i];

	if (strcmp(IMList[i]->objname, objname) == 0)
	    return;
    }

    /* 沒有可用位置, 那就配置一個空間 */
    if (im == NULL)
    {
        NumberOfIM++;
	im = (im_t *)oxim_malloc(sizeof(im_t), True);
	if (NumberOfIM == 1)
	{
	    IMList = (im_t **)oxim_malloc(sizeof(im_t *), True);
	}
	else
	{
	    IMList = (im_t **)oxim_realloc(IMList, NumberOfIM * sizeof(im_t *));
	}
	IMList[NumberOfIM - 1] = im;
    }

    char *cmd[2], value[256];

    cmd[0] = objname;
    cmd[1] = "SETKEY";
    im->modname = strdup(modname);
    im->objname = strdup(objname);
    if (!get_resource(xrc, cmd, value, 256, 2))
    {
	im->key = -1;
    }
    else
    {
	set_data(&im->key, RC_INT, value, 0, 0);
	if (im->key < 0 || im->key > 9)
	    im->key = -1;
    }

    cmd[1] = "ALIAS_NAME";
    if (get_resource(xrc, cmd, value, 256, 2))
    {
	im->inpname = strdup(value);
    }
    else
    {
	im->inpname = strdup((cname ? cname : "unknown"));
    }
}

static int CheckTable(char *path, char *name, char *ret_cname)
{
    char *fullpath = oxim_malloc(BUFLEN, True);
    gzFile *fp;
    int valid = False;

    /* 檔案完整路徑 */
    snprintf(fullpath, BUFLEN, "%s/%s", path, name);

    /* 開檔 */
    if ((fp = gzopen(fullpath, "rb")) != NULL)
    {
    	cintab_head_t *tab_hd;
	char *modID[MODULE_ID_SIZE];
	int hd_size = sizeof(cintab_head_t);
	int size = sizeof(char) * MODULE_ID_SIZE;

	if (gzread(fp , modID, size) == size && strcmp((char *)modID, "gencin") == 0)
	{
	    tab_hd = oxim_malloc(sizeof(cintab_head_t), True);
	    /* 檔頭 */
	    if (gzread(fp, tab_hd, hd_size) == hd_size)
	    {
		if (strcmp(GENCIN_VERSION, tab_hd->version) == 0)
		{
		    strcpy(ret_cname, tab_hd->cname);
		    valid = True;
		}
	    }
	    free(tab_hd);
	}
	gzclose(fp);
    }

    free(fullpath);
    return valid;
}

static void ScanDirRegisterIM(oxim_rc_t *xrc, char *path, char *extname)
{
    struct dirent **namelist;
    struct stat stat;
    const unsigned int ext_len = strlen(extname);
    int n;
    char cname[256];

    if (chdir(path) == 0)
    {
	n = scandir(path, &namelist, 0, 0);
	if (n < 0) /* 錯誤 */
	{
	    perr(OXIMMSG_IWARNING, N_("Scan dir: \"%s\" No any Files!\n"), path);
	}
	else
	{
	    while (n--)
	    {
		char *name = namelist[n]->d_name;
		unsigned int name_len;
		lstat(name, &stat);
		/* 一般檔案或 Ssymbolic link */
		if ((name_len = strlen(name)) > ext_len &&
		    strcmp(name, default_module_name) != 0 &&
		   (S_ISREG(stat.st_mode) || S_ISLNK(stat.st_mode)))
		{
		    /* 是否符合副檔名 */
		    if (strcmp(name + name_len - ext_len, extname) == 0)
		    {
			char objname[256];
			strcpy(objname, name);
			objname[name_len - ext_len] = '\0';

			/* 檢查表格或模組是否正確 */
			if (strcmp(extname, ".tab") == 0)
			{
			    if (CheckTable(path, name, cname))
			    {
				OXIM_IMRegister(xrc, "gen-inp", objname, cname);
			    }
			}

			if (strcmp(extname, ".so") == 0)
			{
			    module_t *modp = (module_t *)load_module(objname, MTYPE_IM, MODULE_VERSION, xrc, "modules");
			    if (modp)
			    {
				OXIM_IMRegister(xrc, objname, objname, modp->module_header.name);
				unload_module((mod_header_t *)modp);
			    }
			}
		    }
		}
		free(namelist[n]);
	    }
	    free(namelist);
	}
    }
    else
    {
	perr(OXIMMSG_IWARNING, N_("Can not change dir to \"%s\"\n"), path);
    }
}

/* 取得輸入法總數 */
int OXIM_MaxIM(void)
{
    return NumberOfIM;
}

/* 以索引取得輸入法顯示名稱與 key*/
im_t *OXIM_GetIMByIndex(int idx)
{
    if (idx < 0 || idx >= NumberOfIM)
	return NULL;

    return IMList[idx];
}

/* 以按鍵碼取得輸入法索引值 */
int OXIM_GetIMIdxByKey(int key)
{
    int idx;
    for (idx = 0 ; idx < NumberOfIM ; idx++)
    {
	if (IMList[idx]->key == key)
	    return idx;
    }
    return -1;
}

void OXIM_LoadIMList(oxim_rc_t *xrc)
{
    char *tmp_path = oxim_malloc(BUFLEN, True);

    OXIM_IMFreeAll();
    snprintf(tmp_path, BUFLEN, "%s/tables", xrc->user_dir);
    ScanDirRegisterIM(xrc, tmp_path, ".tab");

    snprintf(tmp_path, BUFLEN, "%s/tables", xrc->default_dir);
    ScanDirRegisterIM(xrc, tmp_path, ".tab");

    snprintf(tmp_path, BUFLEN, "%s/modules", xrc->user_dir);
    ScanDirRegisterIM(xrc, tmp_path, ".so");

    snprintf(tmp_path, BUFLEN, "%s/modules", xrc->default_dir);
    ScanDirRegisterIM(xrc, tmp_path, ".so");

    free(tmp_path);

    /* 依照按鍵順序，排序輸入法 */
    if (NumberOfIM > 1)
    {
	int i, j;
	im_t *imp;
	for (i = 0; i < NumberOfIM ; i++)
	{
	    for (j = i + 1 ; j < NumberOfIM ; j++)
	    {
		if (IMList[i]->key > IMList[j]->key)
		{
		    imp = IMList[i];
		    IMList[i] = IMList[j];
		    IMList[j] = imp;
		}
	    }
	}
    }
}
