/*
    Copyright (C) 1999 by  XCIN TEAM

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/      

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <locale.h>
#include <unistd.h>
#include "oximtool.h"
#include "gencin.h"
#include "oxim2tab.h"

static ccode_info_t ccinfo;

static cintab_head_t th;
static ichar_t *ichar;
static icode_t *icode1;
static icode_t *icode2;
static int n_ignore;


/*--------------------------------------------------------------------------

	Normal cin2tab functions.

--------------------------------------------------------------------------*/

static void
cin_ename(char *arg, cintab_t *cintab)
{
    if (! arg[0])
	perr(OXIMMSG_ERROR, N_("%s(%d): arguement expected.\n"), 
		cintab->fname_cin, cintab->lineno);
    strncpy(th.ename, arg, CIN_ENAME_LENGTH);
    th.ename[CIN_ENAME_LENGTH - 1] = '\0';
}

static void
cin_cname(char *arg, cintab_t *cintab)
{
    if (! arg[0])
	perr(OXIMMSG_ERROR, N_("%s(%d): arguement expected.\n"),
		cintab->fname_cin, cintab->lineno);
    strncpy(th.cname, arg, CIN_CNAME_LENGTH);
    th.cname[CIN_CNAME_LENGTH - 1] = '\0';
}

static void
cin_selkey(char *arg, cintab_t *cintab)
{
    if (! arg[0])
	perr(OXIMMSG_ERROR, N_("%s(%d): arguement expected.\n"),
		cintab->fname_cin, cintab->lineno);
    th.n_selkey = strlen(arg);
    if (th.n_selkey > SELECT_KEY_LENGTH)
	perr(OXIMMSG_ERROR, N_("%s(%d): too many selection keys defined.\n"),
	     cintab->fname_cin, cintab->lineno);
    strcpy(th.selkey, arg);
}

static void
cin_endkey(char *arg, cintab_t *cintab)
{
    if (! arg[0])
	perr(OXIMMSG_ERROR, N_("%s(%d): arguement expected.\n"),
		cintab->fname_cin, cintab->lineno);
    th.n_endkey = strlen(arg);
    if (th.n_endkey > END_KEY_LENGTH)
	perr(OXIMMSG_ERROR, N_("%s(%d): too many end keys defined.\n"),
	     cintab->fname_cin, cintab->lineno);
    strcpy(th.endkey, arg);
}

static void
cin_keyname(char *arg, cintab_t *cintab)
{
    char cmd1[64], arg1[64];
    int k;

    if (! arg[0] || strcmp(arg, "begin") != 0)
	perr(OXIMMSG_ERROR, N_("%s(%d): arguement \"begin\" expected.\n"),
	     cintab->fname_cin, cintab->lineno);

    while (cmd_arg(cmd1, 64, arg1, 64, NULL)) {
        if (! arg1[0])
	    perr(OXIMMSG_ERROR, N_("%s(%d): arguement expected.\n"),
		cintab->fname_cin, cintab->lineno);
	if (! strcmp("%keyname", cmd1) && ! strcmp("end", arg1))
	    break;

	if (! (k = key2code(cmd1[0])))
	    perr(OXIMMSG_ERROR, N_("%s(%d): illegal key \"%c\" specified.\n"),
		 cintab->fname_cin, cintab->lineno, cmd1[0]);
	if (th.keyname[k].uch)
	    perr(OXIMMSG_ERROR, N_("%s(%d): key \"%c\" is already in used.\n"),
		 cintab->fname_cin, cintab->lineno, cmd1[0]);
	if (! read_hexwch(th.keyname[k].s, arg1))
	    strncpy((char *)th.keyname[k].s, arg1, UCH_SIZE);
	th.n_keyname++;
    }
}

/*------------------------------------------------------------------------*/

typedef struct {
    unsigned int ucs4;
    unsigned int key[2];
} cin_char_t;

static int
icode_cmp(const void *a, const void *b)
{
    cin_char_t *aa=(cin_char_t *)a, *bb=(cin_char_t *)b;

    if (aa->key[0] == bb->key[0]) {
	if (aa->key[1] == bb->key[1])
	    return 0;
	else if (aa->key[1] > bb->key[1])
	    return 1;
	else
	    return -1;
    }
    else if (aa->key[0] > bb->key[0])
	return 1;
    else
	return -1;
}

static void
cin_chardef(char *arg, cintab_t *cintab)
{
    char cmd1[64], arg1[64], arg2[2];
    cin_char_t *cchar=NULL, *cch=NULL;
    int len, ret;
    unsigned int i;

    if (! arg[0] || strcmp(arg, "begin") != 0)
	perr(OXIMMSG_ERROR, N_("%s(%d): arguement \"begin\" expected.\n"),
	     cintab->fname_cin, cintab->lineno);

    ccinfo.total_char = 0;

    while ((ret=cmd_arg(cmd1, 64, arg1, 64, arg2, 2, NULL))) {
        if (! arg1[0])
	    perr(OXIMMSG_ERROR, N_("%s(%d): arguement expected.\n"),
		cintab->fname_cin, cintab->lineno);
	if (! strcmp("%chardef", cmd1) && ! strcmp("end", arg1))
	    break;

	/*
	 *  Fill in the cin_char_t *cch unit.
	 */
	unsigned int ucs4 = 0;
	if ((ucs4 = ccode_to_ucs4(arg1)) == 0) {
	    n_ignore ++;
	    continue;
	}

	th.n_icode ++;
	if (th.n_icode > ccinfo.total_char)
	{
	    if (ccinfo.total_char == 0)
	    {
		cchar = oxim_malloc(sizeof(cin_char_t), 1);
	    }
	    else
	    {
		cchar = oxim_realloc(cchar, (th.n_icode+1) * sizeof(cin_char_t));
	    }
	    cch = (cchar + th.n_icode) - 1;
	    ccinfo.total_char = th.n_icode;
	}

	cch->ucs4 = ucs4;
	keys2codes(cch->key, 2, cmd1);

	th.n_ichar ++;
	cch ++;

	len = strlen(cmd1);
	if (th.n_max_keystroke < len)
	    th.n_max_keystroke = len;
    }

    /*
     *  Determine the memory model.
     */
    ret = (th.n_max_keystroke <= 5) ? ICODE_MODE1 : ICODE_MODE2;
    th.icode_mode = ret;

    /*
     *  Fill in the ichar, icode_idx and icode1, icode2 arrays.
     */
    stable_sort(cchar, th.n_icode, sizeof(cin_char_t), icode_cmp);

    ichar = oxim_malloc(th.n_icode * sizeof(ichar_t), 1);
    icode1 = oxim_malloc(th.n_icode*sizeof(icode_t), 1);
    if (ret == ICODE_MODE2)
        icode2 = oxim_malloc(th.n_icode*sizeof(icode_t), 1);

    for (i=0; i< th.n_icode; i++)
	ichar[i] = 0;
    for (i=0, cch=cchar; i<th.n_icode; i++, cch++) {
	ichar[i] = cch->ucs4;
	icode1[i] = (icode_t)(cch->key[0]);
        if (ret == ICODE_MODE2)
	    icode2[i] = (icode_t)(cch->key[1]);
    }
    free(cchar);
}


/*----------------------------------------------------------------------------

	Main Functions.

----------------------------------------------------------------------------*/

struct genfunc_s {
    char *name;
    void (*func) (char *, cintab_t *);
    byte_t gotit;
};

static struct genfunc_s genfunc[] = {
    {"%ename", cin_ename, 0},
    {"%cname", cin_cname, 0},
    {"%keyname", cin_keyname, 0},
    {"%selkey", cin_selkey, 0},
    {"%endkey", cin_endkey, 1},
    {"%chardef", cin_chardef, 0},
    {NULL, NULL, 0}
};

void
gencin(cintab_t *cintab)
{
    int i;
    char cmd[64], arg[64];
    size_t ret;

    strncpy(th.version, GENCIN_VERSION, VERLEN);
    strncpy(th.encoding, cintab->xrc.locale.encoding, ENCLEN);

    while (cmd_arg(cmd, 64, arg, 64, NULL)) {
	for (i=0; genfunc[i].name != NULL; i++) {
	    if (! strcmp(genfunc[i].name, cmd)) {
		genfunc[i].func(arg, cintab);
		genfunc[i].gotit = 1;
		break;
	    }
	}
	if (genfunc[i].name == NULL)
	    perr(OXIMMSG_ERROR, N_("%s(%d): unknown command: %s.\n"),
		 cintab->fname_cin, cintab->lineno, cmd);
    }

    for (i=0; genfunc[i].name != NULL; i++) {
	if (genfunc[i].gotit == 0)
	    perr(OXIMMSG_ERROR, N_("%s: command \"%s\" not specified.\n"),
		 cintab->fname_cin, genfunc[i].name);
    }
    perr(OXIMMSG_NORMAL,
	N_("number of keyname: %d\n"), th.n_keyname);
    perr(OXIMMSG_NORMAL, 
	N_("max length of keystroke: %d\n"), th.n_max_keystroke);
    perr(OXIMMSG_NORMAL, 
	N_("total char number of this encoding: %d\n"), ccinfo.total_char);
    perr(OXIMMSG_NORMAL, 
	N_("number of char defined: %d\n"), th.n_ichar);
    perr(OXIMMSG_NORMAL, 
	N_("number of keystroke code defined: %d\n"), th.n_icode);
    perr(OXIMMSG_NORMAL, 
	N_("number of defined char ignored: %d\n"), n_ignore);
    perr(OXIMMSG_NORMAL, 
	N_("memory model: %d\n"), th.icode_mode);
    ret = gzwrite(cintab->fw, &th, sizeof(cintab_head_t));
    ret = gzwrite(cintab->fw, icode1, sizeof(icode_t) * th.n_icode);
    ret = gzwrite(cintab->fw, ichar, sizeof(ichar_t) * ccinfo.total_char);
    if (th.icode_mode == ICODE_MODE2)
	ret = gzwrite(cintab->fw, icode2, sizeof(icode_t) * th.n_icode);
}

