/*
    Copyright (C) 1999 by  XCIN TEAM

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/      

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <unistd.h>
#include "oximtool.h"
#include "module.h"
#include "oxim2tab.h"

cintab_t cintab;

struct cin {
    char *name, *modname, *cin_version;
    void (*func) (cintab_t *);
};

/*----------------------------------------------------------------------------

	All the converting procedure functions should registered here.

----------------------------------------------------------------------------*/

#include "gencin.h"
#include "bimscin.h"

int verbose;

extern void gencin(cintab_t *cintab);
extern void bimscin(cintab_t *cintab);

static struct cin cinfunc[] = {
/*  <cin tag>,	 <module ID>,	<cin version>	 <converting func>  */
    {"%gen_inp", "gencin", 	GENCIN_VERSION,  gencin},
    {"%bimspin", "bimscin",	BIMSCIN_VERSION, bimscin},
    {NULL, 	 NULL, 		NULL,		 NULL}
};


/*----------------------------------------------------------------------------

	Cin Reading Functions.

----------------------------------------------------------------------------*/

int
cmd_arg(char *cmd, int cmdlen, ...)
{
    char line[256], *s=line, *arg;
    int arglen, n_read=1;
    va_list list;

    va_start(list, cmdlen);
    if (! get_line(line, 256, cintab.fr, &(cintab.lineno), "#\n"))
	return 0;

    cmd[0] = '\0';
    get_word(&s, cmd, cmdlen, NULL);

    while ((arg = va_arg(list, char *))) {
	arglen = va_arg(list, int);
	if (! get_word(&s, arg, arglen, NULL))
	    break;
	n_read ++;
    }
    return n_read;
}

int
read_hexwch(unsigned char *uch_str, char *arg)
{
    if (arg[0] == '0' && (arg[1] == 'x' || arg[1] == 'X')) {
	char *s = arg+2, tmp[3];
	int i;

	while (*s && isxdigit(*s))
	    s ++;
	if (*s)
	    return 0;

	tmp[2] = '\0';
	for (i=0, s=arg+2; i<UCH_SIZE; i++, s+=2) {
	    if (*s) {
	        tmp[0] = *s;
	        tmp[1] = *(s+1);
	        uch_str[i] = (unsigned char)strtoul(tmp, NULL, 16);
	    }
	    else
		uch_str[i] = (unsigned char)0;
	}
	return 1;
    }
    return 0;
}

char *
turncat_fn(char *fn, char *ext_rm, char *ext_add)
{
    char fn_tmp[256], *s=NULL;

    strncpy(fn_tmp, fn, 254-strlen(ext_add));
    if ((s = strrchr(fn_tmp, '.')) && ! strcmp(s+1, ext_rm)) {
	*s = '\0';
        sprintf(s, ".%s", ext_add);
    }
    else if (! s || strcmp(s+1, ext_add)) {
	strcat(fn_tmp, ".");
        strcat(fn_tmp, ext_add);
    }

    return (char *)strdup(fn_tmp);
}

/*----------------------------------------------------------------------------

	Main Functions.

----------------------------------------------------------------------------*/

void
cin2tab(void)
{
    int i;
    char cmd[64], arg[64], modID[MODULE_ID_SIZE];
    size_t ret;

    if (! (cintab.fr = open_file(cintab.fname_cin, "rt", OXIMMSG_EMPTY))) {
	cintab.fr = open_file(cintab.fname, "rt", OXIMMSG_ERROR);
	free(cintab.fname_cin);
	cintab.fname_cin = cintab.fname;
    }
    cintab.fw = gzopen(cintab.fname_tab, "wb");

    if (cmd_arg(cmd, 64, arg, 64, NULL)) {
	for (i=0; cinfunc[i].name && strcmp(cinfunc[i].name, cmd) != 0; i++);
	if (cinfunc[i].name) {
	    perr(OXIMMSG_NORMAL, 
		N_("cin file: %s, use module: %s version %s.\n"),
		cintab.fname_cin, cinfunc[i].modname, cinfunc[i].cin_version);
	    bzero(modID, MODULE_ID_SIZE);
	    strncpy(modID, cinfunc[i].modname, MODULE_ID_SIZE);
	    ret = gzwrite(cintab.fw, modID, MODULE_ID_SIZE);

	    cinfunc[i].func(&cintab);
	}
	else
	    perr(OXIMMSG_ERROR, N_("no module header name specified.\n"));
    }
    fclose(cintab.fr);
    gzclose(cintab.fw);

}

static void
print_usage(void)
{
    int i;

    perr(OXIMMSG_EMPTY, 
	 N_("OXIM2TAB version (oxim %s)\n"
	    "Usage: oxim2tab [-v] [-r <rcfile>] [-l <encoding>]\n"
	    "               [-o output] <cin_fn>\n\n"
	    "Supported module header names:\n\t"),
	    PACKAGE_VERSION);
    for (i=0; cinfunc[i].name; i++)
	perr(OXIMMSG_EMPTY, "%s, ", cinfunc[i].name);
    perr(OXIMMSG_EMPTY, "\n\n");
}

static void
cin2tab_setlocale(locale_t *locale)
{
    char loc_return[128], enc_return[128];
    int ret;

    if (locale->encoding != NULL)
	return;

    ret = set_lc_ctype("", loc_return, 128, enc_return, 128, OXIMMSG_WARNING);
    if (ret == True) {
	locale->lc_ctype = (char *)strdup(loc_return);
	locale->encoding = (char *)strdup(enc_return);
    }
    else {
	set_lc_ctype_env("", loc_return, 128, enc_return, 128, OXIMMSG_WARNING);
	locale->lc_ctype = (char *)strdup(loc_return);
	locale->encoding = (char *)strdup(loc_return);
    }
}

int
main(int argc, char **argv)
{
    char *s;
    int rev;
#ifdef HPUX
    extern char *optarg;
    extern int opterr, optopt, optind;
#endif

    set_perr("oxim2tab");
    set_lc_messages("", NULL, 0);
    cintab.xrc.argc = argc;
    cintab.xrc.argv = argv;
    if (argc < 2) {
	print_usage();
        exit(0);
    }
    opterr = 0;
    while ((rev = getopt(argc, argv, "v:hr:l:o:")) != EOF) {
        switch (rev) {
	case 'v':
	    verbose = atoi(optarg);
	    break;
	case 'h':
	    print_usage();
	    exit(0);
	case 'l':
	    if ((s = strchr(optarg, '.')) != NULL) {
		cintab.xrc.locale.lc_ctype = (char *)strdup(optarg);
		cintab.xrc.locale.encoding = (char *)strdup(s+1);
	    }
	    else
		cintab.xrc.locale.encoding = (char *)strdup(optarg);
	    s = cintab.xrc.locale.encoding;
	    while (*s) {
		*s = (char)tolower(*s);
		s ++;
	    }
	    break;
	case 'o':
	    cintab.fname_tab = (char *)strdup(optarg);
	    break;
        case '?':
            perr(OXIMMSG_ERROR, N_("unknown option  -%c.\n"), optopt);
            break;
        }
    }
    cin2tab_setlocale(&(cintab.xrc.locale));

    perr(OXIMMSG_EMPTY, N_("OXIM2TAB version (%s) encoding=%s\n"),
	PACKAGE_VERSION, cintab.xrc.locale.encoding);

    if (! argv[optind])
	perr(OXIMMSG_ERROR, N_("no cin file specified.\n"));
    if (! cintab.xrc.locale.encoding)
	perr(OXIMMSG_ERROR, N_("no valid encoding specified.\n"));
    cintab.fname = (char *)strdup(argv[optind]);
    cintab.fname_cin = turncat_fn(argv[optind], "cin", "cin");
    if (! cintab.fname_tab)
	cintab.fname_tab = turncat_fn(argv[optind], "cin", "tab");

    cin2tab();
    return 0;
}
