/*
    Copyright (C) 1999 by  XCIN TEAM

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "config.h"
#include <stdlib.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/xpm.h>
#include <X11/extensions/shape.h>
#include "gui.h"
#include "oxim.h"

typedef struct {
    char *color_name;
    unsigned int color_idx;
    char *color_value;
} color_t;

static color_t colors[] =
{
    {"WINBOX_COLOR", WINBOX_COLOR, "#737173"},
    {"BORDER_COLOR", BORDER_COLOR, "#eeeee6"},
    {"LIGHT_COLOR", LIGHT_COLOR, "#ffffff"},
    {"DARK_COLOR", DARK_COLOR, "#b4b29c"},
    {"CURSOR_COLOR", CURSOR_COLOR, "#7f0000"},
    {"CURSORFONT_COLOR", CURSORFONT_COLOR, "#ffff00"},
    {"FONT_COLOR", FONT_COLOR, "#000000"},
    {"CONVERTNAME_COLOR", CONVERTNAME_COLOR, "#7f0000"},
    {"INPUTNAME_COLOR", INPUTNAME_COLOR, "#00007f"},
    {"UNDERLINE_COLOR", UNDERLINE_COLOR, "black"},
    {"KEYSTROKE_COLOR", KEYSTROKE_COLOR, "#0000ff"},
    {"KEYSTROKEBOX_COLOR", KEYSTROKEBOX_COLOR, "#ffffff"},
    {"SELECTFONT_COLOR", SELECTFONT_COLOR, "#ffffff"},
    {"SELECTBOX_COLOR", SELECTBOX_COLOR, "#3179ac"},
    {"MENU_BG_COLOR", MENU_BG_COLOR, "#4c59a6"},
    {"MENU_FONT_COLOR", MENU_FONT_COLOR, "#ffffff"},
    {"XCIN_BORDER_COLOR", XCIN_BORDER_COLOR, "#ffffff"},
    {"XCIN_BG_COLOR", XCIN_BG_COLOR, "#0000ff"},
    {"XCIN_FONT_COLOR", XCIN_FONT_COLOR, "#ffffff"},
    {"XCIN_STATUS_FG_COLOR", XCIN_STATUS_FG_COLOR, "#00ffff"},
    {"XCIN_STATUS_BG_COLOR", XCIN_STATUS_BG_COLOR, "#000000"},
    {"XCIN_CURSOR_FG_COLOR", XCIN_CURSOR_FG_COLOR, "#ffff00"},
    {"XCIN_CURSOR_BG_COLOR", XCIN_CURSOR_BG_COLOR, "#7f0000"},
    {"XCIN_UNDERLINE_COLOR", XCIN_UNDERLINE_COLOR, "#00ffff"},
    {NULL, 0}
};

Cursor gui_cursor_hand;
Cursor gui_cursor_move;
Cursor gui_cursor_arrow;

winlist_t *gui_root_init(gui_t *gui, xccore_t *xccore);
winlist_t *gui_status_init(gui_t *gui, xccore_t *xccore);
winlist_t *gui_preedit_init(gui_t *gui, xccore_t *xccore);
winlist_t *gui_select_init(gui_t *gui, xccore_t *xccore);
winlist_t *gui_tray_init(gui_t *gui, xccore_t *xccore);
winlist_t *gui_menu_init(gui_t *gui, xccore_t *xccore);
winlist_t *gui_symbol_init(gui_t *gui, xccore_t *xccore);
winlist_t *gui_keyboard_init(gui_t *gui, xccore_t *xccore);
winlist_t *gui_xcin_init(gui_t *gui, xccore_t *xccore);

int check_tray_manager(gui_t *gui);
void xim_terminate(void);
int ic_destroy(int icid, xccore_t *xccore);

static gui_t *gui;

void Xft_Init(xccore_t *xccore);
XftFont *gui_find_font(winlist_t *win, FcChar32 ucs4);
/*----------------------------------------------------------------------------

	Basic WinList handling functions.

----------------------------------------------------------------------------*/

static winlist_t *winlist, *free_win;

static void 
free_win_content(winlist_t *win)
{
    int i;

    if ((win->winmode & WMODE_EXIT))
	return;

    if (win->win_destroy_func)
	win->win_destroy_func(gui, win);
    else {
	XUnmapWindow(gui->display, win->window);
	win->winmode &= ~WMODE_MAP;

	XDestroyWindow(gui->display, win->window);
    }
    win->winmode |= WMODE_EXIT;
    /* Add by Firefly(firefly@ossii.com.tw) */
    XftDrawDestroy(win->draw);
}

winlist_t *
gui_search_win(Window window)
{
    winlist_t *w = winlist;

    while (w) {
	if (w->window == window)
	    return w;
	w = w->next;
    }
    return NULL;
}

winlist_t *
gui_new_win(void)
{
    winlist_t *win;

    if (free_win)
    {
	win = free_win;
	free_win = free_win->next;
	bzero(win, sizeof(winlist_t));
    }
    else
    {
	win = oxim_malloc(sizeof(winlist_t), 1);
    }
    win->win_draw_func = NULL;
    win->win_event_func = NULL;
    win->win_attrib_func = NULL;
    win->win_destroy_func = NULL;
    win->next = winlist;
    winlist = win;

    return win;
}

void
gui_freewin(Window window)
{
    winlist_t *w = winlist, *wp = NULL;

    while (w) {
	if (w->window == window) {
	    free_win_content(w);
	    if (wp == NULL)
		winlist = w->next;
	    else
		wp->next = w->next;
	    w->next = free_win;
	    free_win = w;
	    return;
	}
	wp = w;
	w  = w->next;
    }
}

void
gui_winmap_change(winlist_t *win, int state)
{
    if ((win->winmode & WMODE_EXIT))
	return;

    if (state && !(win->winmode & WMODE_MAP))
    {
	XMapRaised(gui->display, win->window);
	win->winmode |= WMODE_MAP;
    }
    else if (!state && (win->winmode & WMODE_MAP))
    {
	XUnmapWindow(gui->display, win->window);
	win->winmode &= ~WMODE_MAP;
    }
}

int
gui_check_window(Window window)
{
    Window root;
    int x, y;
    unsigned width, height, bw, depth;

    XGetGeometry(gui->display, window,
		&root, &x, &y, &width, &height, &bw, &depth);
    XSync(gui->display, False);
    if (errstatus == 0)
	return True;
    else {
	errstatus = 0;
	return False;
    }
}

int
gui_check_input_focus(xccore_t *xccore, Window win)
{
    Window w;
    int revert_to_return;

    XGetInputFocus(xccore->gui.display, &w, &revert_to_return);
    return (w == win) ? True : False;
}

Window gui_get_input_focus(void)
{
    Window w;
    int rt;
    XGetInputFocus(gui->display, &w, &rt);
    return w;
}

/* 取得 LED 狀態 */
int gui_get_led_state(KeySym keysym)
{
    KeyCode keycode = XKeysymToKeycode(gui->display, keysym);

    if (keycode == NoSymbol)
	return 0;

    int keymask;
    Window w1, w2;
    int n1,n2,n3,n4;
    unsigned int imask;
    XModifierKeymap *map = XGetModifierMapping(gui->display);
    int i;
    for (i = 0 ; i < 8 ; i++)
    {
	if (map->modifiermap[map->max_keypermod * i] == keycode)
	{
	    keymask = 1 << i;
	}
    }

    XQueryPointer(gui->display, gui->root, &w1, &w2,
			&n1, &n2, &n3, &n4, &imask);

    XFreeModifiermap(map);
    return ((keymask & imask) != 0);
}

/*----------------------------------------------------------------------------

	GUI Monitoring the XIM client windows. 

----------------------------------------------------------------------------*/
/*
struct mw_s {
    Window win;
    int icid;
    xmode_t flag;
};

#define MW_LENTH	64
static struct mw_s *mw;
static int mw_edx, mw_len;

void
gui_set_monitor(Window w, xmode_t flag, int icid)
{
    int i, idx=-1;

    for (i=0; i<mw_edx; i++) {
	if (idx == -1 && mw[i].win == (Window)0)
	    idx = i;
	if (mw[i].win == w) {
	    mw[i].flag |= flag;
	    return;
	}
    }
    if (idx == -1) {
	if (mw_edx >= mw_len) {
	    mw_len += MW_LENTH;
	    mw = oxim_realloc(mw, sizeof(struct mw_s)*mw_len);
	    for (i=mw_edx; i<mw_len; i++)
		mw[i].win = (Window)0;
	}
	idx = mw_edx;
	mw_edx ++;
    }
    XSelectInput(gui->display, w, StructureNotifyMask);
    mw[idx].win = w;
    mw[idx].icid = icid;
    mw[idx].flag = flag;
}

static void
gui_unset_monitor(Window w, xccore_t *xccore)
{
    int i;

    for (i=0; i<mw_edx; i++) {
	if (mw[i].win == w &&
	    (mw[i].flag & WIN_MONITOR_CLIENT)) {
	    ic_destroy(mw[i].icid, xccore);
	    mw[i].win = (Window)0;
	    mw[i].icid = 0;
	    mw[i].flag = 0;
	    if (i == mw_edx-1 && mw_edx > 0)
		mw_edx --;
	    break;
	}
    }
}

static void
gui_check_monitor(XConfigureEvent *event)
{
    int i;

    for (i=0; i<mw_edx; i++) {
	if (mw[i].win == event->window &&
	    (mw[i].flag & WIN_MONITOR_OVERSPOT)) {
	    gui_overspot_check_client(gui, mw[i].icid);
	    break;
	}
    }
}
*/
/*----------------------------------------------------------------------------

	GUI Initialization & Main loop

----------------------------------------------------------------------------*/

static int errhandle(Display *disp, XErrorEvent *xevent)
{
    errstatus = xevent->error_code;
    return 1;
}

void
gui_init(xccore_t *xccore)
{
    gui = &(xccore->gui);

    if (! (gui->display = XOpenDisplay(xccore->irc->display_name)))
	perr(OXIMMSG_ERROR, N_("cannot open display: %s\n"), 
			    xccore->irc->display_name);
    (void) XSetErrorHandler(errhandle);
    gui->screen = DefaultScreen(gui->display);
    gui->visual = DefaultVisual(gui->display, gui->screen);
    gui->colormap = DefaultColormap(gui->display, gui->screen);
    gui->display_width = DisplayWidth(gui->display, gui->screen);
    gui->display_height = DisplayHeight(gui->display, gui->screen);
    gui->root = RootWindow(gui->display, gui->screen);
    gui->wm_del_win  = XInternAtom(gui->display, "WM_DELETE_WINDOW", False);

    gui_cursor_hand = XCreateFontCursor(gui->display, XC_hand2);
    gui_cursor_move = XCreateFontCursor(gui->display, XC_fleur);
    gui_cursor_arrow = XCreateFontCursor(gui->display, XC_left_ptr);

    oxim_rc_t *xrc = &(xccore->oxim_rc);
    char *cmd[1], value[256];
    char *color_name, *color_value;
    int color_idx, i;
    for (i = 0 ; i < MAX_COLORS ; i++)
    {
	color_name = colors[i].color_name;
	color_idx = colors[i].color_idx;
	color_value = colors[i].color_value;

	cmd[0] = color_name;
	if (get_resource(xrc, cmd, value, 256, 1))
	    color_value = value;

	XftColorAllocName(gui->display, gui->visual, gui->colormap,
			color_value, &gui->colors[color_idx]);
    }

    Xft_Init(xccore);

    gui->root_win    = gui_root_init(gui, xccore);
    gui->status_win  = gui_status_init(gui, xccore);
    gui->preedit_win = gui_preedit_init(gui, xccore);
    gui->select_win  = gui_select_init(gui, xccore);
    gui->tray_win    = gui_tray_init(gui, xccore);
    gui->menu_win    = gui_menu_init(gui, xccore);
    gui->symbol_win  = gui_symbol_init(gui, xccore);
    gui->keyboard_win  = gui_keyboard_init(gui, xccore);
    gui->xcin_win      = gui_xcin_init(gui, xccore);
    gui->winchange |= WIN_CHANGE_IM;
}

void 
gui_loop(xccore_t *xccore)
{
    static int rx = -1, ry = -1;
    static unsigned int b = 0;

    XEvent event;
    winlist_t *win;

    while (1) {
	if ((xccore->oxim_mode & OXIM_RUN_EXITALL)) {
	    xim_terminate();
	    OXIM_IMFreeAll();
	    exit(0);
	}
#if 0
	{
	    usleep(100);
	    Window focus_win = None;
	    int rt;
	    XGetInputFocus(gui->display, &focus_win, &rt);
	    if (focus_win != None)
	    {
		Window w;
		unsigned int bb = 0;
		int x, y, wx, wy;
		if (XQueryPointer(gui->display, focus_win, &w, &w, &x, &y, &wx, &wy, &bb))
		{
		    if (bb>255 && (bb!=b || rx!=x || ry!=y))
		    {
			b = bb;
			rx = x;
			ry = y;
			if (gui_menu_actived())
			{
			    gui_hide_menu(gui);
			}
		    }
		}
	    }
	}
#endif
	check_tray_manager(gui);

        XNextEvent(gui->display, &event);

	if (XFilterEvent(&event, None) != True)
	{
	    win = gui_search_win(event.xany.window);
	    if (win && win->win_event_func)
		win->win_event_func(gui, win, &(event));
	}
    }
}

/*----------------------------------------------------------------------------

	GUI update window list functions.

----------------------------------------------------------------------------*/

Window
gui_request_init(IM_Context_t *imc, greq_win_t *gw)
{
    Window window;

    switch (gw->greq_data->type) {
	window = (Window)0;
	break;
    }
    return window;
}

void
gui_update_winlist(xccore_t *xccore)
{
    if (gui->xcin_style)
    {
	gui->xcin_win->win_draw_func(gui, gui->xcin_win);
    }
    else
    {
	gui->preedit_win->win_draw_func(gui, gui->preedit_win);
	gui->status_win->win_draw_func(gui, gui->status_win);
    }

    gui->tray_win->win_draw_func(gui, gui->tray_win);

    if (gui->keyboard_win->win_draw_func)
    {
	gui->keyboard_win->win_draw_func(gui, gui->keyboard_win);
    }

    gui->winchange = 0;
}

/*----------------------------------------------------------------------------
    Mouse Event
----------------------------------------------------------------------------*/
static void gui_get_mouse_xy(gui_t *gui, int *x, int *y)
{
    Window ret_root, ret_child;
    int win_x, win_y, mx, my;
    unsigned int mask;

    XQueryPointer(gui->display, gui->root, &ret_root, &ret_child, &win_x, &win_y, &mx, &my, &mask);
    *x = mx;
    *y = my;
}

static void gui_draw_xor_box(gui_t *gui, GC gc, int x, int y, int width, int height)
{
    XSetForeground(gui->display, gc, WhitePixel(gui->display, gui->screen) ^ BlackPixel(gui->display, gui->screen));
    XDrawRectangle(gui->display, gui->root, gc, x, y, width, height);
}

void
gui_move_window(gui_t *gui, winlist_t *win)
{
    int event_x, event_y;
    int offset_x, offset_y;
    int move_x, move_y;
    GC  moveGC;
    int draw_flag = False;

    gui_get_mouse_xy(gui, &event_x, &event_y);
    offset_x = event_x - win->pos_x;
    offset_y = event_y - win->pos_y;

    moveGC = XCreateGC(gui->display, gui->root, 0, NULL);
    XSetSubwindowMode(gui->display, moveGC, IncludeInferiors);
    XSetForeground(gui->display, moveGC, BlackPixel(gui->display, gui->screen));
    XSetFunction(gui->display, moveGC, GXxor);

    XChangeActivePointerGrab(gui->display,
                PointerMotionMask | ButtonMotionMask | ButtonReleaseMask |
                OwnerGrabButtonMask, None, CurrentTime);

    XGrabServer(gui->display);

    XEvent myevent;
    while(1)
    {
	XNextEvent(gui->display, &myevent);
	switch(myevent.type)
	{
	    case ButtonRelease:
		if(myevent.xbutton.button == Button1)
		{
		    if (draw_flag)
			gui_draw_xor_box(gui, moveGC, move_x - offset_x, move_y - offset_y, win->width, win->height);
		    XFreeGC(gui->display, moveGC);
		    gui_get_mouse_xy(gui, &move_x, &move_y);
		    win->pos_x = move_x - offset_x;
		    win->pos_y = move_y - offset_y;

		    /* */
		    if (win->pos_x < 0)
			win->pos_x = 0;
		    if (win->pos_y < 0)
			win->pos_y = 0;
		    if (win->pos_x + win->width > gui->display_width)
			win->pos_x = gui->display_width - win->width;
		    if (win->pos_y + win->height > gui->display_height)
			win->pos_y = gui->display_height - win->height;

		    XMoveWindow(gui->display, win->window,
				 win->pos_x, win->pos_y);
		    XUngrabServer(gui->display);
		    return;
		}
		break;

	    case MotionNotify:
		if (draw_flag)
		    gui_draw_xor_box(gui, moveGC, move_x - offset_x, move_y - offset_y, win->width, win->height);
		gui_get_mouse_xy(gui, &move_x, &move_y);
		gui_draw_xor_box(gui, moveGC, move_x - offset_x, move_y - offset_y, win->width, win->height);
		draw_flag = True;
		break;
	    default:
		break;
	}
    }
}

/******************************************************************
*
* Add by Firefly(firefly@ossii.com.tw)
*
******************************************************************/
#include <X11/Xft/Xft.h>

/* 預設的字型名稱 */
static char default_font_name[256];

static FcFontSet *default_fontset = NULL;

static unsigned int num_fonts = 0;
static XftFont **xftfonts = NULL;

static unsigned int default_fontsize = 0;
static unsigned int leading = 1;

void Xft_Init(xccore_t *xccore)
{
    char *cmd[1], value[256];
    cmd[0] = "DEFAULT_FONT_NAME";
    if (get_resource(&xccore->oxim_rc, cmd, value, 256, 1))
        set_data(default_font_name, RC_STRARR, value, 0, 256);
    else
	strcpy(default_font_name, "serif");

    cmd[0] = "DEFAULT_FONT_SIZE";
    if (get_resource(&xccore->oxim_rc, cmd, value, 256, 1))
	set_data(&default_fontsize, RC_INT, value, 0, 0);

    if (default_fontsize < 12 || default_fontsize > 48)
	default_fontsize = 16;

    XftFontSet *fonts;
    char *familyName;
    int weight, slant, scalable;
    FcCharSet *charset;
    int i;

    default_fontset = FcFontSetCreate();

    fonts = XftListFonts(gui->display, gui->screen, (char *)0,
			XFT_FILE, XFT_INDEX,
			XFT_FAMILY, XFT_WEIGHT, XFT_SLANT, XFT_SPACING,
			FC_CHARSET, FC_SCALABLE,
			(char *)0);

    for (i = 0; i < fonts->nfont; i++)
    {
	FcPattern *pat = fonts->fonts[i];
	if (XftPatternGetString(pat, XFT_FAMILY, 0, &familyName) != XftResultMatch ||
	    XftPatternGetInteger(pat, XFT_WEIGHT, 0, &weight)!= XftResultMatch ||
	    XftPatternGetInteger(pat, XFT_SLANT, 0, &slant)!= XftResultMatch ||
	    XftPatternGetBool(pat, XFT_SCALABLE, 0, &scalable)!= XftResultMatch ||
	    FcPatternGetCharSet(pat, FC_CHARSET, 0, &charset)!= XftResultMatch)
	{
	    continue;
	}

	if (weight > FC_WEIGHT_MEDIUM || slant != FC_SLANT_ROMAN || !scalable)
	{
	    continue;
	}

	FcPattern *dup_pat = FcPatternDuplicate(pat);
	if (dup_pat)
	    FcFontSetAdd(default_fontset, dup_pat);
    }
    XftFontSetDestroy (fonts);

    num_fonts = 1;
    xftfonts = (XftFont **)oxim_malloc(sizeof(XftFont *), True);
    xftfonts[0] = XftFontOpen(gui->display, gui->screen,
		XFT_FAMILY, XftTypeString, default_font_name,
		XFT_PIXEL_SIZE, XftTypeDouble, (double)default_fontsize,
		NULL);
}

/* 計算字串繪圖寬度 */
unsigned int Utf8_Char_Escapement(XftDraw *draw, XftFont *font, char *Utf8Char, int Utf8len)
{
    XGlyphInfo GlyphInfo;
    XftTextExtentsUtf8(XftDrawDisplay(draw), font, (FcChar8 *)Utf8Char, Utf8len,  &GlyphInfo);
    return GlyphInfo.xOff;
}

/* 繪出Utf8字元 */
unsigned int Draw_Utf8_Char(XftDraw *draw, XftFont *font, int foreground_idx, int background_idx, int x, int y, char *Utf8Char, int Utf8len)
{
    unsigned int width = Utf8_Char_Escapement(draw, font, Utf8Char, Utf8len);
    if (background_idx >= 0 && background_idx <= MAX_COLORS)
    {
	XftDrawRect(draw, &gui->colors[background_idx], x, y - default_fontsize+1, width, default_fontsize);
    }

    XftDrawStringUtf8(draw, &gui->colors[foreground_idx], font, x, y, (FcChar8 *)Utf8Char, Utf8len);
    return width;
}

XftFont *gui_find_font(winlist_t *win, FcChar32 ucs4)
{
    unsigned int i;
    unsigned int w_num_fonts;
    XftFont **w_xftfonts;

    /* 如果視窗字型大小與預設相同，用預設就好 */
    if (gui_FontSize(win) == default_fontsize)
    {
	w_num_fonts = num_fonts;
	w_xftfonts  = xftfonts;
    }
    else /* 否則取用該視窗自己的 */
    {
	/* 該視窗沒有預設字型，就從系統預設抓取 */
	if (!win->num_fonts)
	{
	    win->num_fonts = 1;
	    win->xftfonts = (XftFont **)oxim_malloc(sizeof(XftFont *), True);

	    FcPattern *dup_pat = FcPatternDuplicate(xftfonts[0]->pattern);
	    /* 刪除預設 size */
	    XftPatternDel(dup_pat, XFT_PIXEL_SIZE);
	    /* 指定自己的 Size */
	    XftPatternAddInteger(dup_pat, XFT_PIXEL_SIZE, gui_FontSize(win));
	    win->xftfonts[0] = XftFontOpenPattern(gui->display, dup_pat);
	}
	w_num_fonts = win->num_fonts;
	w_xftfonts  = win->xftfonts;
    }

    /* */
    for (i=0 ; i < w_num_fonts ; i++)
    {
	if (XftCharExists(gui->display, w_xftfonts[i], ucs4))
	{
	    return w_xftfonts[i];
	}
    }

    for (i=0; i< default_fontset->nfont; i++)
    {
	FcPattern *pat = default_fontset->fonts[i];
	FcCharSet *fcs = NULL;

	if (FcPatternGetCharSet(pat, FC_CHARSET, 0, &fcs) != XftResultMatch)
	    continue;

	if (!FcCharSetHasChar(fcs, ucs4))
	    continue;

	FcResult res;
	FcPattern *mpat = FcFontMatch(0, pat, &res);
	if (!mpat)
	    continue;

	XftPatternAddInteger(mpat, XFT_PIXEL_SIZE, gui_FontSize(win));
	XftFont *chkfont = XftFontOpenPattern(gui->display, mpat);

        if (chkfont)
	{
	    const unsigned int nfonts = w_num_fonts + 1;
	    w_xftfonts = (XftFont **)oxim_realloc(w_xftfonts, nfonts * sizeof(XftFont *));
	    if (!w_xftfonts)
	    {
		FcPatternDestroy(mpat);
		continue;
	    }
	    w_xftfonts[w_num_fonts] = chkfont;

	    if (gui_FontSize(win) != default_fontsize)
	    {
		win->num_fonts = nfonts;
		win->xftfonts = w_xftfonts;
	    }
	    else
	    {
		num_fonts = nfonts;
	        xftfonts = w_xftfonts;
	    }

	    return chkfont;
	}
	else
	{
	    FcPatternDestroy(mpat);
	}
    }

    return w_xftfonts[0];
}

unsigned int gui_Draw_String(winlist_t *win, int foreground_idx, int background_idx, int x, int y, char *string, int len)
{
    char *p = string;
    unsigned int ucs4;
    unsigned int string_width = 0;
    int nbytes;

    while (len && (nbytes = oxim_utf8_to_ucs4(p, &ucs4, len)) > 0)
    {
        unsigned int width;

	XftFont *font = gui_find_font(win, ucs4);

	width = Draw_Utf8_Char(win->draw, font, foreground_idx, background_idx, x, y, p, nbytes);

	x += width;
	string_width += width;
	p += nbytes;
	len -= nbytes;
    }
    return string_width;
}

unsigned int gui_TextEscapement(winlist_t *win, char *string, int len)
{
    char *p = string;
    unsigned int ucs4;
    unsigned int string_width = 0;
    int nbytes = oxim_utf8_to_ucs4(p, &ucs4, len);

    while (len && (nbytes = oxim_utf8_to_ucs4(p, &ucs4, len)) > 0)
    {
	XftFont *font = gui_find_font(win, ucs4);

	unsigned int width = Utf8_Char_Escapement(win->draw, font, p, nbytes);

	string_width += width;
	p += nbytes;
	len -= nbytes;
    }

    return string_width;
}

void gui_Draw_3D_Box(winlist_t *win, int x, int y, int width, int height, int fill_idx, int boxstyle)
{
    GC lightGC = XCreateGC(gui->display, win->window, 0, NULL);
    GC darkGC  = XCreateGC(gui->display, win->window, 0, NULL);

    if (boxstyle)
    {
	XSetForeground(gui->display, lightGC, gui_Color(LIGHT_COLOR));
	XSetForeground(gui->display, darkGC, gui_Color(DARK_COLOR));
    }
    else
    {
	XSetForeground(gui->display, lightGC, gui_Color(DARK_COLOR));
	XSetForeground(gui->display, darkGC, gui_Color(LIGHT_COLOR));
    }
    XSetLineAttributes(gui->display, lightGC, 1, LineSolid, CapRound, JoinRound);
    XSetLineAttributes(gui->display, darkGC, 1, LineSolid, CapRound, JoinRound);
    
    int x2 = x + width - 1;
    int y2 = y + height - 1;
    XDrawLine(gui->display, win->window, lightGC, x, y, x2, y);
    XDrawLine(gui->display, win->window, lightGC, x, y, x, y2);
    XDrawLine(gui->display, win->window, darkGC, x, y2, x2, y2);
    XDrawLine(gui->display, win->window, darkGC, x2, y, x2, y2);
    if (fill_idx >= 0 && fill_idx <= MAX_COLORS)
    {
	XftDrawRect(win->draw, &gui->colors[fill_idx], x+1, y + 1, width-2, height-2);
    }
    XFreeGC(gui->display, lightGC);
    XFreeGC(gui->display, darkGC);
}

void gui_Draw_Line(winlist_t *win, int x, int y, int x1, int y1, int color_idx, int line_style)
{
    XGCValues gv;

    gv.line_width = 1;
    gv.dashes = 1;
    gv.line_style = (line_style) ? LineSolid : LineOnOffDash;

    GC lineGC = XCreateGC(gui->display, win->window, GCLineWidth|GCLineStyle|GCDashList, &gv); 
    XSetForeground(gui->display, lineGC, gui_Color(color_idx));
    XDrawLine(gui->display, win->window, lineGC, x, y, x1, y1);
    XFreeGC(gui->display, lineGC);
}

unsigned long gui_Color(int color_index)
{
    return gui->colors[color_index].pixel;
}

unsigned int gui_FontSize(winlist_t *win)
{

    if (win->font_size)
	return win->font_size;

    return default_fontsize;
}

void gui_Draw_Image(winlist_t *win, int x, int y, char **xpm_data)
{
    XImage* image;

    XpmCreateImageFromData(gui->display, xpm_data, &image, NULL, NULL);
    XPutImage(gui->display, win->window, DefaultGC(gui->display, gui->screen),
                image, 0, 0, x, y, image->width, image->height);
    XDestroyImage(image);
}

/* 送出模擬按鍵到指定的視窗 */
void gui_send_key(Window win, int state, KeySym keysym)
{
    XEvent Keyevent;

    Keyevent.type = KeyPress;
    Keyevent.xany.type = KeyPress;
    Keyevent.xany.display = gui->display;
    Keyevent.xkey.serial = 0L;
    Keyevent.xkey.send_event = True;
    Keyevent.xkey.window = win;
    Keyevent.xkey.subwindow = None;
    Keyevent.xkey.root = gui->root;
    Keyevent.xkey.time = CurrentTime;
    Keyevent.xkey.x = 0;
    Keyevent.xkey.y = 0;
    Keyevent.xkey.x_root = 0;
    Keyevent.xkey.y_root = 0;
    Keyevent.xkey.same_screen = True;
    Keyevent.xkey.keycode = XKeysymToKeycode(gui->display, keysym);
    Keyevent.xkey.state = state;

    XSendEvent(gui->display, win, False, KeyPressMask, (XEvent *)&Keyevent);
}
