/*
    Copyright (C) 2006 by OXIM TEAM

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include <stdlib.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/xpm.h>
#include <string.h>
#include "oximtool.h"
#include "gui.h"
#include "oxim.h"

#include "images/halfchar.xpm"
#include "images/fullchar.xpm"
#include "images/out_cn.xpm"
#include "images/out_tw.xpm"
#include "images/sideback_l.xpm"
#include "images/sideback.xpm"
#include "images/sideback_r.xpm"

#define X_LEADING       2
#define Y_LEADING       2

static void 
win_draw(gui_t *gui, winlist_t *win, IC *ic)
{
    char *inpn=NULL;
    char **inpb=NULL, **inp_conv=NULL;
    char *s, buf[100];

    IM_Context_t *imc = ic->imc;
    int x, y, len, img_y;
    img_y = (win->height - 16) / 2;
    inpb = ((imc->inp_state & IM_2BYTES)) ? icon_full : icon_half;

    if (imc->inp_state & IM_OUTSIMP)
	inp_conv = icon_cn;
    else if (imc->inp_state & IM_OUTTRAD)
	inp_conv = icon_tw;

    if ((imc->inp_state & IM_CINPUT)) {
        s = imc->inpinfo.inp_cname;
        while (*s) {
            if (*s == '%' && *(s+1) == '%' && *(s+2)) {
                inpn = s+2;
                break;
            }
            s++;
        }
        if (! inpn) {
            strcpy((char *)buf, imc->inpinfo.inp_cname);
            inpn = buf;
        }
    }
    else
        inpn = "En";

    /* TODO: 如果沒有變動的話，就不必重新繪製 status win
    */

    len = strlen((char *)inpn);

    gui_winmap_change(win, 1);
    XClearWindow(gui->display, win->window);
    
    gui_Draw_Image(win, 0, 0, sideback_l);
    x = 15 ;
    y = win->height - ((win->height - win->font_size) / 2) - 1;

    int inpn_w = gui_TextEscapement(win, inpn, len);

    gui_Draw_String(win, FONT_COLOR, NO_COLOR, x+1, y+1, inpn, len);
    x += gui_Draw_String(win, LIGHT_COLOR, NO_COLOR, x, y, inpn, len);

    x += (X_LEADING * 2) + 1;
    gui_Draw_Image(win, x, img_y, inpb);
    x += 16 + (X_LEADING * 2);

    if (inp_conv)
    {
	gui_Draw_Image(win, x, img_y, inp_conv);
	x += 20;
    }

    win->width = x + 5;
    gui_Draw_Image(win, win->width - 5, 0, sideback_r);

    if (win->width + win->pos_x > gui->display_width)
    {
	win->pos_x = gui->display_width - win->width;
	XMoveWindow(gui->display, win->window, win->pos_x, win->pos_y);
    }
    XResizeWindow(gui->display, win->window, win->width, win->height);
}

static void
gui_status_draw(gui_t *gui, winlist_t *win)
{
    xccore_t *xccore = (xccore_t *)win->data;
    IC *ic = xccore->ic;
    inp_state_t inp_state = (ic) ? ic->imc->inp_state : 0;
    if (((win->winmode & WMODE_EXIT)) || !ic || (!(inp_state & IM_CINPUT) && !(inp_state & IM_2BYTES)) || gui->xcin_style) {
	gui_winmap_change(win, 0);
	return;
    }
    win_draw(gui, win, ic);
}

static void StatusEventProcess(gui_t *gui, winlist_t *win, XEvent *event)
{
    switch (event->type)
    {
	case Expose:
	    if (event->xexpose.count == 0)
	    {
		gui_status_draw(gui, win);
	    }
	    break;
	case GraphicsExpose:
	    gui_status_draw(gui, win);
	    break;
	case MotionNotify: /* Mouse Move */
	    if (event->xbutton.x < 12)
	    {
		XDefineCursor(gui->display, win->window, gui_cursor_move);
	    }
	    else
	    {
		XDefineCursor(gui->display, win->window, gui_cursor_arrow);
	    }
	    break;
	case ButtonPress:
	    if (event->xbutton.button == Button1 && event->xbutton.x < 12)
	    {
		gui_move_window(gui, win);
	    }
	    break;
    }
}

winlist_t *
gui_status_init(gui_t *gui, xccore_t *xccore)
{
    char *cmd[1], value[256];
    unsigned int font_size = 0;
    cmd[0] = "STATUS_FONT_SIZE";
    if (get_resource(&xccore->oxim_rc, cmd, value, 256, 1))
        set_data(&font_size, RC_INT, value, 0, 0);

    if (font_size < 12 || font_size > 18)
        font_size = 16;

    winlist_t *win = gui_new_win();
    win->width    = 1;
    win->height   = 24;
    win->font_size = font_size;
    win->pos_x = gui->display_width - 200;
    win->pos_y = gui->display_height - 64;
    win->window = XCreateSimpleWindow(gui->display, gui->root, 
		win->pos_x, win->pos_y, win->width, win->height, 0, 0, 0);
    win->draw = XftDrawCreate(gui->display, win->window,
                DefaultVisual(gui->display, gui->screen),
                gui->colormap);

    win->data = (void *)xccore;
    win->win_draw_func = gui_status_draw;
    win->win_event_func = StatusEventProcess;

    XSelectInput(gui->display, win->window, (ExposureMask|ButtonPressMask|ButtonReleaseMask|StructureNotifyMask|PointerMotionMask));

    Pixmap pix;
    XpmCreatePixmapFromData(gui->display, win->window, sideback, &pix, NULL, NULL);
    XSetWindowAttributes win_attr;
    win_attr.background_pixmap = pix;
    win_attr.backing_store = Always;
    win_attr.override_redirect = True;
    win_attr.save_under = True;
    XChangeWindowAttributes(gui->display, win->window,
			CWBackPixmap|CWBackingStore|CWOverrideRedirect|CWSaveUnder, &win_attr);
    XFreePixmap(gui->display, pix);

    return win;
}
