/*
    Copyright (C) 2006 by OXIM TEAM

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/Xatom.h>
#include <string.h>
#include "oximtool.h"
#include "oxim.h"

#define X_LEADING	2
#define Y_LEADING	2

#define DRAW_EMPTY	1
#define	DRAW_MCCH	2
#define	DRAW_PRE_MCCH	3
#define DRAW_LCCH	4

#define XIMPreeditMask	0x00ffL
#define XIMStatusMask	0xff00L

#define FLAG_KEYSTROKE_ONSPOT	0x0001L
#define FLAG_KEYSTROKE_OVERSPOT	0x0002L
#define FLAG_KEYSTROKE_ROOT	0x0004L
#define FLAG_KEYSTROKE_MASK	0x000fL

#define FLAG_LCCH_ONSPOT	0x0010L
#define FLAG_LCCH_OVERSPOT	0x0020L
#define FLAG_LCCH_ROOT		0x0040L
#define FLAG_LCCH_MASK		0x00f0L

#define FLAG_MCCH_ONBOX		0x0100L
#define FLAG_MCCH_ROOT		0x0200L
#define FLAG_MCCH_MASK		0x0f00L

#define BUFSIZE 256

extern XIMS ims;

static void 
check_has_set_spot(gui_t *gui, IC *ic)
{
    if (ic->has_set_spot || ic->never_set_spot)
	return;

    Window w;
    w = (ic->ic_rec.focus_win) ? ic->ic_rec.focus_win : ic->ic_rec.client_win;

    Atom spot_prop = XInternAtom(gui->display, "ONTHESPOT_LOCATION", False);
    Atom ret_type;
    int ret_format;
    unsigned long ret_len, ret_after;
    unsigned char *ret_property;

    if (XGetWindowProperty(gui->display, w, spot_prop, 0, sizeof(XPoint), False, XA_POINT, &ret_type, &ret_format, &ret_len, &ret_after, &ret_property) == Success && (ret_type == XA_POINT))
    {
	memcpy(&ic->ic_rec.pre_attr.spot_location, ret_property, sizeof(XPoint));
	return;
    }

    ic->never_set_spot = True;
}

static char
*mbstocts(gui_t *gui, char *s)
{
    XTextProperty tp;
    //XmbTextListToTextProperty(gui->display, &s, 1, XCompoundTextStyle, &tp);
    Xutf8TextListToTextProperty(gui->display, &s, 1, XCompoundTextStyle, &tp);
    return (char *)tp.value;
}

static int draw_multich(gui_t *gui, IC *ic, int x, int y, int preview) 
{
    int i, j, n_groups, n, x2, len=0, toggle_flag;
    uch_t *selkey, *cch;
    inpinfo_t inpinfo = ic->imc->inpinfo;
    winlist_t *win = gui->xcin_win;

    if (inpinfo.n_mcch == 0)
	return x;

    cch = inpinfo.mcch;
    selkey = inpinfo.s_selkey;
    if (!inpinfo.mcch_grouping || inpinfo.mcch_grouping[0]==0) {
	toggle_flag = 0;
	n_groups = inpinfo.n_mcch;
    }
    else {
	toggle_flag = 1;
	n_groups = inpinfo.mcch_grouping[0];
    }

    for (i=0; i<n_groups && toggle_flag!=-1; i++, selkey++)
    {
	n = (toggle_flag > 0) ? inpinfo.mcch_grouping[i+1] : 1;

	/* 顯示選擇按鍵 */
	if (selkey->uch != (uchar_t)0)
	{
	    len = strlen((char *)selkey->s);
	    x += gui_Draw_String(win, XCIN_CURSOR_FG_COLOR, XCIN_CURSOR_BG_COLOR, x, y, (char *)selkey->s, len) + X_LEADING;
	}

	/* 顯示候選字詞 */
	for (j=0; j<n; j++, cch++)
	{
	    len = strlen((char *)cch->s);
	    if (cch->uch == (uchar_t)0)
	    {
		toggle_flag = -1;
		break;
	    }
	    x2 = x + gui_TextEscapement(win, (char *)cch->s, len);
	    if (inpinfo.mcch_hint && inpinfo.mcch_hint[(int)(cch-inpinfo.mcch)])
	    {
		gui_Draw_Line(win, x, y+2, x2, y+2, XCIN_UNDERLINE_COLOR, 1);
	    }
	    gui_Draw_String(win, XCIN_FONT_COLOR, XCIN_BG_COLOR, x, y, (char *)cch->s, len);
	    x = x2;
	}
	x += (X_LEADING * 2);
    }

    if (! (inpinfo.guimode & GUIMOD_SELKEYSPOT))
	return x;

    char *pgstate = NULL;
    switch (inpinfo.mcch_pgstate)
    {
	case MCCH_BEGIN:
	    pgstate = "<";
	    break;
	case MCCH_MIDDLE:
	    pgstate = "<>";
	    break;
	case MCCH_END:
	    pgstate = ">";
	    break;
    }

    if (pgstate)
    {
	x += gui_Draw_String(win, XCIN_FONT_COLOR, XCIN_BG_COLOR, x, y, pgstate, strlen(pgstate)) + (X_LEADING * 2);
    }

    return x;
}

static void draw_keystroke_onspot(gui_t *gui, IC *ic)
{
    if (!(ic->ic_rec.input_style & XIMPreeditCallbacks))
	return;

    uch_t *keystroke = ic->imc->inpinfo.s_keystroke;

    if (!keystroke || keystroke[0].uch==(uchar_t)0)
        return;

    char buf[256];
    int len = wchs_to_mbs(buf, keystroke, 255);

    XIMText text;
    IMPreeditCBStruct data;
    XIMFeedback feedback[256];
    int i;
    char *tmp;
    for (i=0 ; i < ic->imc->inpinfo.keystroke_len ; i++)
	feedback[i] = XIMReverse;

    data.major_code = XIM_PREEDIT_DRAW;
    data.connect_id = ic->connect_id;
    data.icid = ic->id;
    data.todo.draw.caret = ic->preedit_length;
    data.todo.draw.chg_first = ic->preedit_length;
    data.todo.draw.chg_length = ic->imc->inpinfo.keystroke_len;
    data.todo.draw.text = &text;
    text.encoding_is_wchar = False;
    text.feedback = feedback;

    {
	tmp = mbstocts(gui, buf);
	text.string.multi_byte = tmp;
	text.length = strlen(tmp);
	feedback[text.length] = 0;
	IMCallCallback(ims, (XPointer)&data);
	XFree(tmp);
	ic->preedit_length += ic->imc->inpinfo.keystroke_len;
    }
}

static void
draw_preedit_onspot(gui_t *gui, IC *ic)
{
    char buf[BUFSIZE];

    inpinfo_t inpinfo = ic->imc->inpinfo;
    int len = wchs_to_mbs(buf, inpinfo.lcch, BUFSIZE);

    if (!len && !ic->preedit_length)
	return;

    int i;
    XIMFeedback feedback[BUFSIZE] = {0};
    for (i=0 ; i <= inpinfo.n_lcch ; i++)
	feedback[i] = XIMUnderline;

    if (inpinfo.n_lcch && inpinfo.edit_pos < inpinfo.n_lcch)
	feedback[inpinfo.edit_pos] |= XIMReverse;

    XIMText text;
    IMPreeditCBStruct data;
    char *tmp;

    data.major_code = XIM_PREEDIT_DRAW;
    data.connect_id = ic->connect_id;
    data.icid = ic->id;
    data.todo.draw.chg_first = 0;
    data.todo.draw.chg_length = ic->preedit_length;
    data.todo.draw.text = &text;
    text.encoding_is_wchar = False;
    text.feedback = feedback;

    if (!len)
    {
	xim_preeditcallback_done(ic);
    }
    else
    {
	tmp = mbstocts(gui, buf);
	text.string.multi_byte = tmp;
	text.length = strlen(tmp);
        data.todo.draw.caret = (inpinfo.edit_pos < inpinfo.n_lcch) ? inpinfo.edit_pos : inpinfo.n_lcch;
	IMCallCallback(ims, (XPointer)&data);
	XFree(tmp);
	ic->preedit_length = inpinfo.n_lcch;
    }
}

static int
draw_preedit_overspot(gui_t *gui, winlist_t *win, int x, int y)
{
    xccore_t *xccore = (xccore_t *)win->data;
    IC *ic = xccore->ic;
    int len;
    char buf[BUFSIZE];
    inpinfo_t inpinfo = ic->imc->inpinfo;

    if (inpinfo.n_lcch == 0)
	return x;

    /* 字詞群組 */
    if (inpinfo.lcch_grouping)
    {
	int i, x2, y2, n_cch=0, n_seg, save_x = x;
	ubyte_t *glist = inpinfo.lcch_grouping+1;
	char str[128];
	for (i = 0; i < inpinfo.lcch_grouping[0]; i++)
	{
	    n_seg = glist[i];
	    len = nwchs_to_mbs(str, inpinfo.lcch+n_cch, n_seg, 128);
	    x2 = x + gui_TextEscapement(win, str, len);
	    if (n_seg > 1)
	    {
		int gx1 = x + (X_LEADING * 2);
		int gy1 = y + 2;
		int gx2 = x2 -(X_LEADING * 2);
		int gy2 = y + 4;
		gui_Draw_Line(win, gx1, gy1, gx1, gy2, XCIN_UNDERLINE_COLOR, 1);
		gui_Draw_Line(win, gx1, gy2, gx2, gy2, XCIN_UNDERLINE_COLOR, 1);
		gui_Draw_Line(win, gx2, gy1, gx2, gy2+1, XCIN_UNDERLINE_COLOR, 1);
	    }
	    x = x2;
	    n_cch += n_seg;
	}
	x = save_x;
    }

    len = wchs_to_mbs(buf, inpinfo.lcch, BUFSIZE);
    int lcch_w = gui_TextEscapement(win, buf, len);
    int cursor_w = gui_TextEscapement(win, "  ", 2);

    if (inpinfo.edit_pos < inpinfo.n_lcch) {
	uch_t *tmp = inpinfo.lcch + inpinfo.edit_pos;

	if (inpinfo.edit_pos > 0) {
	    len = nwchs_to_mbs(buf, inpinfo.lcch, inpinfo.edit_pos, BUFSIZE);
	    x += gui_Draw_String(win, XCIN_FONT_COLOR, XCIN_BG_COLOR, x, y, buf, len);
	}
	len = strlen((char *)tmp->s);
	x += gui_Draw_String(win, XCIN_CURSOR_FG_COLOR, XCIN_CURSOR_BG_COLOR, x, y, (char *)tmp->s, len);
        if (inpinfo.edit_pos < inpinfo.n_lcch - 1) {
            len = wchs_to_mbs(buf, inpinfo.lcch + inpinfo.edit_pos+1, BUFSIZE);
	    x += gui_Draw_String(win, XCIN_FONT_COLOR, XCIN_BG_COLOR, x, y, buf, len);
        }
    }
    else {
        len = wchs_to_mbs(buf, inpinfo.lcch, BUFSIZE);
        if (len)
	    x += gui_Draw_String(win, XCIN_FONT_COLOR, XCIN_BG_COLOR, x, y, buf, len);
        else
            x = X_LEADING;
	x += gui_Draw_String(win, XCIN_FONT_COLOR, XCIN_CURSOR_BG_COLOR, x, y, "  ", 2);
    }
    return x + (X_LEADING*2);
}

static int
draw_keystroke_overspot(gui_t *gui, winlist_t *win, int x, int y)
{
    xccore_t *xccore = (xccore_t *)win->data;
    IC *ic = xccore->ic;
    uch_t *keystroke = ic->imc->inpinfo.s_keystroke;

    if (!keystroke || keystroke[0].uch==(uchar_t)0)
	return x;

    char buf[256];
    buf[0] = '['; 
    int len = wchs_to_mbs(buf + 1, keystroke, 254);
    buf[len+1] = ']';
    buf[len+2] = '\0';

    x += gui_Draw_String(win, XCIN_FONT_COLOR, XCIN_BG_COLOR, x, y, buf, len+2) + (X_LEADING * 2);

    return x;
}

static int draw_inpname(gui_t *gui, winlist_t *win, int x, int y)
{
    xccore_t *xccore = (xccore_t *)win->data;
    IC *ic = xccore->ic;
    IM_Context_t *imc = ic->imc;
    char inpname[100], *inpn=NULL, *inpb=NULL, *inp_conv=NULL;
    char *s, buf[100];
    int len;

    if (imc->inp_state & IM_OUTSIMP)
        inp_conv = "出简";
    else if (imc->inp_state & IM_OUTTRAD)
        inp_conv = "出繁";

    inpb = ((imc->inp_state & IM_2BYTES)) ? "全形" : "半形";
    if ((imc->inp_state & IM_CINPUT))
    {
	strcpy(buf, imc->inpinfo.inp_cname);
	inpn = buf;
    }
    else
    {
	inpn = "英數";
    }

    if (inp_conv)
	len = snprintf(inpname, 100, "%s|%s|%s", inpn, inpb, inp_conv);
    else
	len = snprintf(inpname, 100, "%s|%s", inpn, inpb);

    x += gui_Draw_String(win, XCIN_STATUS_FG_COLOR, XCIN_STATUS_BG_COLOR, x, y, inpname, len) + (X_LEADING * 3);
    return x;
}

static int draw_cch_publish(gui_t *gui, winlist_t *win, int x, int y)
{
    xccore_t *xccore = (xccore_t *)win->data;
    IC *ic = xccore->ic;
    IM_Context_t *imc = ic->imc;

    if (imc->inpinfo.cch_publish.uch)
    {
	char *str, buf[128];
	int len;
	len = sprintf(buf, "%s:", (char *)imc->inpinfo.cch_publish.s);
	str = buf + len;
	if (imc->sinmd_keystroke[0].uch &&
	    wchs_to_mbs(str, imc->sinmd_keystroke, 128-len))
	{
	    len = strlen(buf);
	    x += gui_Draw_String(win, XCIN_FONT_COLOR, XCIN_BG_COLOR, x, y, buf, len) + (X_LEADING * 2);
	}
    }
    return x;
}

static void xcin_win_draw(gui_t *gui, winlist_t *win, unsigned int flag)
{
    xccore_t *xccore = (xccore_t *)win->data;
    IC *ic = xccore->ic;
    IM_Context_t *imc = ic->imc;

    Window w, junkwin;
    int ret_x, ret_y;
    int x = X_LEADING * 2;
    int y = win->height - Y_LEADING - 5 - 1 ;
    w = (ic->ic_rec.focus_win) ? ic->ic_rec.focus_win : ic->ic_rec.client_win;
    XTranslateCoordinates(gui->display, w, gui->root, 0, 0, &ret_x, &ret_y, &junkwin);
    check_has_set_spot(gui, ic);
    x = draw_inpname(gui, win, x, y);

    if (flag)
    {
	/* 判斷預編輯文字 */
	switch (flag & FLAG_LCCH_MASK)
	{
	    case FLAG_LCCH_ONSPOT: /* 預編輯文字在游標之上 */
		draw_preedit_onspot(gui, ic);
		break;
	    case FLAG_LCCH_OVERSPOT:
	    case FLAG_LCCH_ROOT:
		x = draw_preedit_overspot(gui, win, x, y);
		break;
	}

	switch (flag & FLAG_KEYSTROKE_MASK)
	{
	    case FLAG_KEYSTROKE_ONSPOT:
		draw_keystroke_onspot(gui, ic);
		break;
	    case FLAG_KEYSTROKE_OVERSPOT:
	    case FLAG_KEYSTROKE_ROOT:
		x = draw_keystroke_overspot(gui, win, x, y);
		break;
	}

	int preview = (imc->inpinfo.guimode & GUIMOD_SELKEYSPOT) ? False : True;
	x = draw_multich(gui, ic, x, y, preview);
    }
    else
    {
	x = draw_cch_publish(gui, win, x, y);
    }

    int new_x = ret_x + ic->ic_rec.pre_attr.spot_location.x + 2;
    int new_y = ret_y + ic->ic_rec.pre_attr.spot_location.y + 2;
    if (new_x + x > gui->display_width)
	new_x = gui->display_width - x;

    if (x != win->width)
    {
	win->width = x;
	XResizeWindow(gui->display, win->window, win->width, win->height);
    }

    if (!(ic->never_set_spot) && (new_x != win->pos_x || new_y != win->pos_y))
    {
	win->pos_x = new_x;
	win->pos_y = new_y;
	XMoveWindow(gui->display, win->window, new_x, new_y);
    }

}

static void gui_xcin_draw(gui_t *gui, winlist_t *win)
{
    xccore_t *xccore = (xccore_t *)win->data;
    IC *ic = xccore->ic;
    unsigned long flag = 0;

    if ((win->winmode & WMODE_EXIT) || ic == NULL || !gui->xcin_style)
	return;

    IM_Context_t *imc = ic->imc;

    if (!(imc->inp_state & IM_XIMFOCUS) && !(imc->inp_state & IM_2BFOCUS))
    {
	gui_winmap_change(win, False);
	return;
    }

    inpinfo_t inpinfo = imc->inpinfo;

    switch (ic->ic_rec.input_style & XIMPreeditMask)
    {
	case XIMPreeditCallbacks: /* On The Spot */
	    if (inpinfo.keystroke_len)
	    {
		/* 無法取得游標位置 */
		if (ic->never_set_spot)
		    flag |= FLAG_KEYSTROKE_ONSPOT;
		else
		    flag |= FLAG_KEYSTROKE_OVERSPOT;
	    }
	    if (inpinfo.n_lcch || ic->preedit_length)
		flag |= FLAG_LCCH_ONSPOT;
	    if (inpinfo.n_mcch)
	    {
		if (!(inpinfo.guimode & GUIMOD_SELKEYSPOT) || ic->never_set_spot)
		    flag |= FLAG_MCCH_ONBOX;
		else
		    flag = FLAG_MCCH_ONBOX;
	    }
	    break;
	case XIMPreeditPosition:  /* Over The Spot */
	    if (inpinfo.keystroke_len)
		flag |= FLAG_KEYSTROKE_OVERSPOT;
	    if (inpinfo.n_lcch)
		flag |= FLAG_LCCH_OVERSPOT;
	    if (inpinfo.n_mcch)
	    {
		if (imc->inpinfo.guimode & GUIMOD_SELKEYSPOT)
		    flag = FLAG_MCCH_ONBOX;
		else
		    flag |= FLAG_MCCH_ONBOX;
	    }
	    break;
	case XIMPreeditArea:	  /* Off The Spot */
	    break;
	case XIMPreeditNone:	/* Root Window */
	case XIMPreeditNothing:
	    if (inpinfo.keystroke_len)
		flag |= FLAG_KEYSTROKE_ROOT;
	    if (inpinfo.n_lcch)
		flag |= FLAG_LCCH_ROOT;
	    if (inpinfo.n_mcch)
		flag |= FLAG_MCCH_ROOT;
	    break;
    }
    gui_winmap_change(win, True);
    xcin_win_draw(gui, win, flag);
}

static void XCINEventProcess(gui_t *gui, winlist_t *win, XEvent *event)
{
    XDefineCursor(gui->display, win->window, gui_cursor_move);
    switch (event->type)
    {
	case Expose:
	    if (event->xexpose.count == 0)
	    {
		gui_xcin_draw(gui, win);
	    }
	    break;
	case GraphicsExpose:
	    gui_xcin_draw(gui, win);
	    break;
	case ButtonPress:
	    gui_move_window(gui, win);
	    break;
    }
}

void gui_xcin_style_switch(gui_t *gui)
{
    gui->xcin_style = !gui->xcin_style;
    if (gui->xcin_style)
    {
        gui_winmap_change(gui->preedit_win, False);
        gui_winmap_change(gui->select_win, False);
        gui_winmap_change(gui->status_win, False);
    }
    else
    {
        gui_winmap_change(gui->xcin_win, False);
    }
}

/* xcin 輸入風格視窗 */
winlist_t *
gui_xcin_init(gui_t *gui, xccore_t *xccore)
{
    oxim_rc_t *xrc = &xccore->oxim_rc;
    winlist_t *win = gui_new_win();

    char *cmd[1], value[256];
    unsigned int font_size = 0;
    cmd[0] = "XCIN_FONT_SIZE";
    if (get_resource(xrc, cmd, value, 256, 1))
        set_data(&font_size, RC_INT, value, 0, 0);

    if (font_size < 12 || font_size > 48)
        font_size = 16;

    win->font_size = font_size;
    win->pos_x = gui->display_width / 2;
    win->pos_y = gui->display_height / 2;
    win->width  = X_LEADING * 2;
    win->height = (Y_LEADING * 2) + font_size + 10;
    win->window = XCreateSimpleWindow(gui->display, gui->root,
                win->pos_x, win->pos_y, win->width, win->height, 1,
                gui_Color(XCIN_BORDER_COLOR), gui_Color(XCIN_BG_COLOR));
    win->draw = XftDrawCreate(gui->display, win->window,
                DefaultVisual(gui->display, gui->screen),
                gui->colormap);

    win->data = (void *)xccore;
    win->win_draw_func  = gui_xcin_draw;
    win->win_event_func = XCINEventProcess;

    XSetWindowAttributes win_attr;
    win_attr.override_redirect = True;
    XChangeWindowAttributes(gui->display, win->window,
                        CWOverrideRedirect, &win_attr);
    XSelectInput(gui->display, win->window, (ExposureMask|ButtonPressMask|ButtonReleaseMask|PointerMotionMask|LeaveWindowMask|StructureNotifyMask));

    return win;
}
